"""
Unit tests for Distance module
"""
# Info
__author__ = 'Jason Anthony Vander Heiden'
from changeo import __version__, __date__

# Imports
import os
import time
import unittest
import numpy as np
from Bio.Seq import Seq

# Presto and changeo imports
import changeo.Distance as Distance

# Paths
test_path = os.path.dirname(os.path.realpath(__file__))
data_path = os.path.join(test_path, 'data')


class Test_Distance(unittest.TestCase):
    def setUp(self):
        print('-> %s()' % self._testMethodName)

        self.nt_seq = ['TGTGCAAGGGGGCCA',
                       'TGTGCAAGGGGGCCA',
                       'TGTATTTGGGGGCCA',
                       'ACACTTGCCACTGAT',
                       'NNNNNNNNNNNNTGA',
                       'NNNNNNNNNNNNNNN']
        self.nt_seq_short = ['TGTGCAAGG',
                             'TGTGCAAGG',
                             'TGTATTTGG',
                             'NNNNNNNNN']
        self.aa_seq = [str(Seq(x).translate()) for x in self.nt_seq]


        # Amino acid distance from sequence 0 to sequence 0,..,5
        self.aa_len = 5.0
        self.aa_ham = np.array([0.0, 0.0, 2.0, 5.0, 1.0, 0.0])

        # Nucleotide distances sequence 0 to sequence 0,..,5
        self.nt_len = 15.0
        # 0 vs 2
        #   A-C = 0; A-G = 1; A-T = 2;
        #   C-G = 0; C-T = 1; G-T = 0
        # 0 vs 3
        #   A-C = 1; A-G = 2; A-T = 4;
        #   C-G = 6; C-T = 1; G-T = 1
        # 0 vs 4
        #   A-C = 0; A-G = 0; A-T = 0;
        #   C-G = 1; C-T = 1; G-T = 0
        self.nt_ham = np.array([0.0, 0.0, 4.0, 15.0, 2.0, 0.0])
        self.nt_hh_s1f = np.array([0.0,
                                   0.0,
                                   0.0 + 1*0.64 + 2*1.16 + 0.0 + 1*0.64 + 0.0,
                                   1*1.21 + 2*0.64 + 4*1.16 + 6*1.16 + 1*0.64 + 1*1.21,
                                   0.0 + 0.0 + 0.0 + 0.0 + 1*1.16 + 1*0.64 + 0.0,
                                   0.0])
        self.nt_mk_rs1nf = np.array([0.0,
                                     0.0,
                                     0.0 + 1*0.32 + 2*1.17 + 0.0 + 1*0.32 + 0.0,
                                     1*1.51 + 2*0.32 + 4*1.17 + 6*1.17 + 1*0.32 + 1*1.51,
                                     0.0 + 0.0 + 0.0 + 0.0 + 1*1.17 + 1*0.32 + 0.0,
                                     0.0])

        # 5-mer models use shorter sequence
        # For sequence 0 to sequence 0,..,3
        self.nt_short_len = 9.0
        # hh_hs5f
        #   GTGCA-GTATT = [0.97, 0.84]
        #   TGCAA-TATTT = [0.93, 0.83]
        #   GCAAG-ATTTG = [1.08, 1.16]
        #   CAAGG-TTTGG = [0.91, 1.07]
        self.nt_hh_s5f_avg = np.array([0.0,
                                       0.0,
                                       np.mean([0.97, 0.84]) + np.mean([0.93, 0.83]) +
                                       np.mean([1.08, 1.16]) + np.mean([0.91, 1.07]),
                                       0.0])
        self.nt_hh_s5f_min = np.array([0.0,
                                       0.0,
                                       np.min([0.97, 0.84]) + np.min([0.93, 0.83]) +
                                       np.min([1.08, 1.16]) + np.min([0.91, 1.07]),
                                       0.0])
        # mk_rs5nf
        #   GTGCA-GTATT = [0.71, 0.77]
        #   TGCAA-TATTT = [0.71, 0.93]
        #   GCAAG-ATTTG = [1.05, 1.03]
        #   CAAGG-TTTGG = [1.08, 1.13]
        self.nt_mk_rs5nf_avg = np.array([0.0,
                                         0.0,
                                         np.mean([0.71, 0.77]) + np.mean([0.71, 0.93]) +
                                         np.mean([1.05, 1.03]) + np.mean([1.08, 1.13]),
                                         0.0])
        self.nt_mk_rs5nf_min = np.array([0.0,
                                         0.0,
                                         np.min([0.71, 0.77]) + np.min([0.71, 0.93]) +
                                         np.min([1.05, 1.03]) + np.min([1.08, 1.13]),
                                         0.0])

        self.start = time.time()

    def tearDown(self):
        t = time.time() - self.start
        print("<- %s() %.3f" % (self._testMethodName, t))
        
    def test_calcDistances(self):
        # aa
        print(' AA>', self.aa_seq)
        result = Distance.calcDistances(self.aa_seq, n=1, dist_mat=Distance.aa_model, norm=None)
        print(' AA>', result[1, :])
        np.testing.assert_almost_equal(result[1, :], self.aa_ham)

        result = Distance.calcDistances(self.aa_seq, n=1, dist_mat=Distance.aa_model, norm='len')
        np.testing.assert_almost_equal(result[1, :], self.aa_ham / self.aa_len)

        # ham
        print(' NT>', self.nt_seq)
        result = Distance.calcDistances(self.nt_seq, n=1, dist_mat=Distance.ham_model, norm=None)
        print('HAM>', result[1, :])
        np.testing.assert_almost_equal(result[1, :], self.nt_ham)

        result = Distance.calcDistances(self.nt_seq, n=1, dist_mat=Distance.ham_model, norm='len')
        np.testing.assert_almost_equal(result[1, :], self.nt_ham / self.nt_len)

        # hh_s1f
        result = Distance.calcDistances(self.nt_seq, n=1, dist_mat=Distance.hh_s1f_model, norm=None)
        print('HH_S1F>\n', result[1, :])
        np.testing.assert_almost_equal(result[1, :], self.nt_hh_s1f)

        result = Distance.calcDistances(self.nt_seq, n=1, dist_mat=Distance.hh_s1f_model, norm='len')
        np.testing.assert_almost_equal(result[1, :], self.nt_hh_s1f / self.nt_len)

        # mk_rs1nf
        result = Distance.calcDistances(self.nt_seq, n=1, dist_mat=Distance.mk_rs1nf_model, norm=None)
        print('MK_RS1NF>\n', result[1, :])
        np.testing.assert_almost_equal(result[1, :], self.nt_mk_rs1nf)

        result = Distance.calcDistances(self.nt_seq, n=1, dist_mat=Distance.mk_rs1nf_model, norm='len')
        np.testing.assert_almost_equal(result[1, :], self.nt_mk_rs1nf / self.nt_len)

        # hh_s5f
        result = Distance.calcDistances(self.nt_seq_short, n=5, dist_mat=Distance.hh_s5f_model, sym='avg', norm=None)
        print('HH_S5F-AVG>\n', result[1, :])
        np.testing.assert_almost_equal(result[1, :], self.nt_hh_s5f_avg)

        result = Distance.calcDistances(self.nt_seq_short, n=5, dist_mat=Distance.hh_s5f_model, sym='avg', norm='len')
        np.testing.assert_almost_equal(result[1, :], self.nt_hh_s5f_avg / self.nt_short_len)

        result = Distance.calcDistances(self.nt_seq_short, n=5, dist_mat=Distance.hh_s5f_model, sym='min', norm=None)
        print('HH_S5F-MIN>\n', result[1, :])
        np.testing.assert_almost_equal(result[1, :], self.nt_hh_s5f_min)

        result = Distance.calcDistances(self.nt_seq_short, n=5, dist_mat=Distance.hh_s5f_model, sym='min', norm='len')
        np.testing.assert_almost_equal(result[1, :], self.nt_hh_s5f_min / self.nt_short_len)

        # mk_rs5nf
        result = Distance.calcDistances(self.nt_seq_short, n=5, dist_mat=Distance.mk_rs5nf_model, sym='avg', norm=None)
        print('MK_RS5NF-AVG>\n', result[1, :])
        np.testing.assert_almost_equal(result[1, :], self.nt_mk_rs5nf_avg)

        result = Distance.calcDistances(self.nt_seq_short, n=5, dist_mat=Distance.mk_rs5nf_model, sym='avg', norm='len')
        np.testing.assert_almost_equal(result[1, :], self.nt_mk_rs5nf_avg / self.nt_short_len)

        result = Distance.calcDistances(self.nt_seq_short, n=5, dist_mat=Distance.mk_rs5nf_model, sym='min', norm=None)
        print('MK_RS5NF-MIN>\n', result[1, :])
        np.testing.assert_almost_equal(result[1, :], self.nt_mk_rs5nf_min)

        result = Distance.calcDistances(self.nt_seq_short, n=5, dist_mat=Distance.mk_rs5nf_model, sym='min', norm='len')
        np.testing.assert_almost_equal(result[1, :], self.nt_mk_rs5nf_min / self.nt_short_len)

        # Raise error on length mismatch
        with self.assertRaises(IndexError):
            Distance.calcDistances(['ATGCAA', 'ATGC'], n=1, dist_mat=Distance.ham_model)


    @unittest.skip("-> loadModels() skipped\n")
    def test_loadModels(self):
        print('MK_RS1NF> ', Distance.mk_rs1nf_model)
        print('  HH_S1F> ', Distance.hh_s5f_model)

        self.fail('TODO')


if __name__ == '__main__':
    unittest.main()



"""
Receptor data structure
"""

# Info
__author__ = 'Jason Anthony Vander Heiden, Namita Gupta, Scott Christley'

# Imports
from collections import OrderedDict
from Bio.Seq import Seq
# import yaml
# from pkg_resources import resource_stream

# Presto and changeo imports
from presto.IO import printError, printWarning
from changeo.Gene import getAllele, getGene, getFamily, getAlleleNumber

# class Schema:
#     """
#     Schema for mapping Receptor attributes to column names
#     """
#     def __init__(self, schema):
#         """
#         Initializer
#
#         Arguments:
#           schema (str): name of schema to load.
#
#         Returns:
#           changeo.Receptor.Schema
#         """
#         with resource_stream(__name__, 'data/receptor.yaml') as f:
#             data = yaml.load(f, Loader=yaml.FullLoader)
#             receptor = {v[schema]: k for k, v in data['receptor'].items()}
#             definition = data[schema]
#
#         # Output extension
#         self.out_type = definition['out_type']
#
#         # Field sets
#         self.fields = list(receptor.keys())
#         self.required = definition['standard']
#         self.custom_fields = definition['custom']
#
#         # Mapping of schema column names to Receptor attributes
#         self._schema_map = {k : receptor[k] for k in self.fields}
#         self._receptor_map = {v: k for k, v in self._schema_map.items()}
#
#     def toReceptor(self, field):
#         """
#         Returns a Receptor attribute name from an Schema column name
#
#         Arguments:
#           field (str): schema column name.
#         Returns:
#           str: Receptor attribute name.
#         """
#         return self._schema_map.get(field, field)
#
#     def fromReceptor(self, field):
#         """
#         Returns a schema column name from a Receptor attribute name
#
#         Arguments:
#           field (str): Receptor attribute name.
#
#         Returns:
#           str: schema column name.
#         """
#         return self._receptor_map.get(field, field)
#
#
# AIRRSchema = Schema('airr')
# ChangeoSchema = Schema('changeo')


class AIRRSchema:
    """
    AIRR format to Receptor mappings
    """
    # Default file extension
    out_type = 'tsv'

    # Core fields
    required = ['sequence_id',
                'sequence',
                'sequence_alignment',
                'germline_alignment',
                'rev_comp',
                'productive',
                'stop_codon',
                'vj_in_frame',
                'locus',
                'v_call',
                'd_call',
                'j_call',
                'c_call',
                'junction',
                'junction_length',
                'junction_aa',
                'np1_length',
                'np2_length',
                'v_sequence_start',
                'v_sequence_end',
                'v_germline_start',
                'v_germline_end',
                'd_sequence_start',
                'd_sequence_end',
                'd_germline_start',
                'd_germline_end',
                'j_sequence_start',
                'j_sequence_end',
                'j_germline_start',
                'j_germline_end']

    # Mapping of AIRR column names to Receptor attributes
    _schema_map = OrderedDict([('sequence_id', 'sequence_id'),
                               ('sequence', 'sequence_input'),
                               ('sequence_alignment', 'sequence_imgt'),
                               ('germline_alignment', 'germline_imgt'),
                               ('sequence_aa', 'sequence_aa_input'),
                               ('sequence_aa_alignment', 'sequence_aa_imgt'),
                               ('germline_aa_alignment', 'germline_aa_imgt'),
                               ('rev_comp', 'rev_comp'),
                               ('productive', 'functional'),
                               ('stop_codon', 'stop'),
                               ('vj_in_frame', 'in_frame'),
                               ('v_frameshift', 'v_frameshift'),
                               ('locus', 'locus'),
                               ('v_call', 'v_call'),
                               ('d_call', 'd_call'),
                               ('j_call', 'j_call'),
                               ('c_call', 'c_call'),
                               ('junction', 'junction'),
                               ('junction_start', 'junction_start'),
                               ('junction_end', 'junction_end'),
                               ('junction_length', 'junction_length'),
                               ('junction_aa', 'junction_aa'),
                               ('junction_aa_length', 'junction_aa_length'),
                               ('np1_length', 'np1_length'),
                               ('np2_length', 'np2_length'),
                               ('np1_aa_length', 'np1_aa_length'),
                               ('np2_aa_length', 'np2_aa_length'),
                               ('v_sequence_start', 'v_seq_start'),
                               ('v_sequence_end', 'v_seq_end'),
                               ('v_sequence_length', 'v_seq_length'),
                               ('v_germline_start', 'v_germ_start_imgt'),
                               ('v_germline_end', 'v_germ_end_imgt'),
                               ('v_germline_length', 'v_germ_length_imgt'),
                               ('v_sequence_aa_start', 'v_seq_aa_start'),
                               ('v_sequence_aa_end', 'v_seq_aa_end'),
                               ('v_sequence_aa_length', 'v_seq_aa_length'),
                               ('v_germline_aa_start', 'v_germ_aa_start_imgt'),
                               ('v_germline_aa_end', 'v_germ_aa_end_imgt'),
                               ('v_germline_aa_length', 'v_germ_aa_length_imgt'),
                               ('d_sequence_start', 'd_seq_start'),
                               ('d_sequence_end', 'd_seq_end'),
                               ('d_sequence_length', 'd_seq_length'),
                               ('d_germline_start', 'd_germ_start'),
                               ('d_germline_end', 'd_germ_end'),
                               ('d_germline_length', 'd_germ_length'),
                               ('d_sequence_aa_start', 'd_seq_aa_start'),
                               ('d_sequence_aa_end', 'd_seq_aa_end'),
                               ('d_sequence_aa_length', 'd_seq_aa_length'),
                               ('d_germline_aa_start', 'd_germ_aa_start'),
                               ('d_germline_aa_end', 'd_germ_aa_end'),
                               ('d_germline_aa_length', 'd_germ_aa_length'),
                               ('j_sequence_start', 'j_seq_start'),
                               ('j_sequence_end', 'j_seq_end'),
                               ('j_sequence_length', 'j_seq_length'),
                               ('j_germline_start', 'j_germ_start'),
                               ('j_germline_end', 'j_germ_end'),
                               ('j_germline_length', 'j_germ_length'),
                               ('j_sequence_aa_start', 'j_seq_aa_start'),
                               ('j_sequence_aa_end', 'j_seq_aa_end'),
                               ('j_sequence_aa_length', 'j_seq_aa_length'),
                               ('j_germline_aa_start', 'j_germ_aa_start'),
                               ('j_germline_aa_end', 'j_germ_aa_end'),
                               ('j_germline_aa_length', 'j_germ_aa_length'),
                               ('germline_alignment_d_mask', 'germline_imgt_d_mask'),
                               ('v_score', 'v_score'),
                               ('v_identity', 'v_identity'),
                               ('v_support', 'v_evalue'),
                               ('v_cigar', 'v_cigar'),
                               ('d_score', 'd_score'),
                               ('d_identity', 'd_identity'),
                               ('d_support', 'd_evalue'),
                               ('d_cigar', 'd_cigar'),
                               ('j_score', 'j_score'),
                               ('j_identity', 'j_identity'),
                               ('j_support', 'j_evalue'),
                               ('j_cigar', 'j_cigar'),
                               ('vdj_score', 'vdj_score'),
                               ('cdr1', 'cdr1_imgt'),
                               ('cdr2', 'cdr2_imgt'),
                               ('cdr3', 'cdr3_imgt'),
                               ('fwr1', 'fwr1_imgt'),
                               ('fwr2', 'fwr2_imgt'),
                               ('fwr3', 'fwr3_imgt'),
                               ('fwr4', 'fwr4_imgt'),
                               ('cdr1_aa', 'cdr1_aa_imgt'),
                               ('cdr2_aa', 'cdr2_aa_imgt'),
                               ('cdr3_aa', 'cdr3_aa_imgt'),
                               ('fwr1_aa', 'fwr1_aa_imgt'),
                               ('fwr2_aa', 'fwr2_aa_imgt'),
                               ('fwr3_aa', 'fwr3_aa_imgt'),
                               ('fwr4_aa', 'fwr4_aa_imgt'),
                               ('cdr1_start', 'cdr1_start'),
                               ('cdr1_end', 'cdr1_end'),
                               ('cdr2_start', 'cdr2_start'),
                               ('cdr2_end', 'cdr2_end'),
                               ('cdr3_start', 'cdr3_start'),
                               ('cdr3_end', 'cdr3_end'),
                               ('fwr1_start', 'fwr1_start'),
                               ('fwr1_end', 'fwr1_end'),
                               ('fwr2_start', 'fwr2_start'),
                               ('fwr2_end', 'fwr2_end'),
                               ('fwr3_start', 'fwr3_start'),
                               ('fwr3_end', 'fwr3_end'),
                               ('fwr4_start', 'fwr4_start'),
                               ('fwr4_end', 'fwr4_end'),
                               ('n1_length', 'n1_length'),
                               ('n2_length', 'n2_length'),
                               ('p3v_length', 'p3v_length'),
                               ('p5d_length', 'p5d_length'),
                               ('p3d_length', 'p3d_length'),
                               ('p5j_length', 'p5j_length'),
                               ('d_frame', 'd_frame'),
                               ('cdr3_igblast', 'cdr3_igblast'),
                               ('cdr3_igblast_aa', 'cdr3_igblast_aa'),
                               ('duplicate_count', 'dupcount'),
                               ('consensus_count', 'conscount'),
                               ('umi_count', 'umicount'),
                               ('clone_id', 'clone'),
                               ('cell_id', 'cell')])

    # Mapping of Receptor attributes to AIRR column names
    _receptor_map = {v: k for k, v in _schema_map.items()}

    # All fields
    fields = list(_schema_map.keys())

    @staticmethod
    def toReceptor(field):
        """
        Returns a Receptor attribute name from an AIRR column name

        Arguments:
          field : AIRR column name.
        Returns:
          str: Receptor attribute name.
        """
        field = field.lower()
        return AIRRSchema._schema_map.get(field, field)

    @staticmethod
    def fromReceptor(field):
        """
        Returns an AIRR column name from a Receptor attribute name

        Arguments:
          field : Receptor attribute name.

        Returns:
          str: AIRR column name.
        """
        field = field.lower()
        return AIRRSchema._receptor_map.get(field, field)


class AIRRSchemaAA(AIRRSchema):
    """
    AIRR format to Receptor amino acid mappings
    """
    # Core fields
    required = ['sequence_id',
                'sequence',
                'sequence_alignment',
                'germline_alignment',
                'sequence_aa',
                'sequence_aa_alignment',
                'germline_aa_alignment',
                'rev_comp',
                'productive',
                'stop_codon',
                'locus',
                'v_call',
                'd_call',
                'j_call',
                'c_call',
                'junction',
                'junction_length',
                'junction_aa',
                'v_sequence_aa_start',
                'v_sequence_aa_end',
                'v_germline_aa_start',
                'v_germline_aa_end']


class ChangeoSchema:
    """
    Change-O to Receptor mappings
    """
    # Default file extension
    out_type = 'tab'

    # Standard fields
    required = ['SEQUENCE_ID',
                'SEQUENCE_INPUT',
                'FUNCTIONAL',
                'IN_FRAME',
                'STOP',
                'MUTATED_INVARIANT',
                'INDELS',
                'LOCUS',
                'V_CALL',
                'D_CALL',
                'J_CALL',
                'SEQUENCE_VDJ',
                'SEQUENCE_IMGT',
                'V_SEQ_START',
                'V_SEQ_LENGTH',
                'V_GERM_START_VDJ',
                'V_GERM_LENGTH_VDJ',
                'V_GERM_START_IMGT',
                'V_GERM_LENGTH_IMGT',
                'NP1_LENGTH',
                'D_SEQ_START',
                'D_SEQ_LENGTH',
                'D_GERM_START',
                'D_GERM_LENGTH',
                'NP2_LENGTH',
                'J_SEQ_START',
                'J_SEQ_LENGTH',
                'J_GERM_START',
                'J_GERM_LENGTH',
                'JUNCTION',
                'JUNCTION_LENGTH',
                'GERMLINE_IMGT']

    # Mapping of Change-O column names to Receptor attributes
    _schema_map = OrderedDict([('SEQUENCE_ID', 'sequence_id'),
                               ('SEQUENCE_INPUT', 'sequence_input'),
                               ('SEQUENCE_AA_INPUT', 'sequence_aa_input'),
                               ('FUNCTIONAL', 'functional'),
                               ('IN_FRAME', 'in_frame'),
                               ('STOP', 'stop'),
                               ('MUTATED_INVARIANT', 'mutated_invariant'),
                               ('INDELS', 'indels'),
                               ('V_FRAMESHIFT', 'v_frameshift'),
                               ('LOCUS', 'locus'),
                               ('V_CALL', 'v_call'),
                               ('D_CALL', 'd_call'),
                               ('J_CALL', 'j_call'),
                               ('C_CALL', 'c_call'),
                               ('SEQUENCE_VDJ', 'sequence_vdj'),
                               ('SEQUENCE_IMGT', 'sequence_imgt'),
                               ('SEQUENCE_AA_VDJ', 'sequence_aa_vdj'),
                               ('SEQUENCE_AA_IMGT', 'sequence_aa_imgt'),
                               ('V_SEQ_START', 'v_seq_start'),
                               ('V_SEQ_LENGTH', 'v_seq_length'),
                               ('V_GERM_START_VDJ', 'v_germ_start_vdj'),
                               ('V_GERM_LENGTH_VDJ', 'v_germ_length_vdj'),
                               ('V_GERM_START_IMGT', 'v_germ_start_imgt'),
                               ('V_GERM_LENGTH_IMGT', 'v_germ_length_imgt'),
                               ('V_SEQ_AA_START', 'v_seq_aa_start'),
                               ('V_SEQ_AA_LENGTH', 'v_seq_aa_length'),
                               ('V_GERM_AA_START_VDJ', 'v_germ_aa_start_vdj'),
                               ('V_GERM_AA_LENGTH_VDJ', 'v_germ_aa_length_vdj'),
                               ('V_GERM_AA_START_IMGT', 'v_germ_aa_start_imgt'),
                               ('V_GERM_AA_LENGTH_IMGT', 'v_germ_aa_length_imgt'),
                               ('NP1_LENGTH', 'np1_length'),
                               ('NP1_AA_LENGTH', 'np1_aa_length'),
                               ('D_SEQ_START', 'd_seq_start'),
                               ('D_SEQ_LENGTH', 'd_seq_length'),
                               ('D_GERM_START', 'd_germ_start'),
                               ('D_GERM_LENGTH', 'd_germ_length'),
                               ('D_SEQ_AA_START', 'd_seq_aa_start'),
                               ('D_SEQ_AA_LENGTH', 'd_seq_aa_length'),
                               ('D_GERM_AA_START', 'd_germ_aa_start'),
                               ('D_GERM_AA_LENGTH', 'd_germ_aa_length'),
                               ('NP2_LENGTH', 'np2_length'),
                               ('NP2_AA_LENGTH', 'np2_aa_length'),
                               ('J_SEQ_START', 'j_seq_start'),
                               ('J_SEQ_LENGTH', 'j_seq_length'),
                               ('J_GERM_START', 'j_germ_start'),
                               ('J_GERM_LENGTH', 'j_germ_length'),
                               ('J_SEQ_AA_START', 'j_seq_aa_start'),
                               ('J_SEQ_AA_LENGTH', 'j_seq_aa_length'),
                               ('J_GERM_AA_START', 'j_germ_aa_start'),
                               ('J_GERM_AA_LENGTH', 'j_germ_aa_length'),
                               ('JUNCTION', 'junction'),
                               ('JUNCTION_LENGTH', 'junction_length'),
                               ('GERMLINE_IMGT', 'germline_imgt'),
                               ('GERMLINE_AA_IMGT', 'germline_aa_imgt'),
                               ('JUNCTION_START', 'junction_start'),
                               ('V_SCORE', 'v_score'),
                               ('V_IDENTITY', 'v_identity'),
                               ('V_EVALUE', 'v_evalue'),
                               ('V_BTOP', 'v_btop'),
                               ('V_CIGAR', 'v_cigar'),
                               ('D_SCORE', 'd_score'),
                               ('D_IDENTITY', 'd_identity'),
                               ('D_EVALUE', 'd_evalue'),
                               ('D_BTOP', 'd_btop'),
                               ('D_CIGAR', 'd_cigar'),
                               ('J_SCORE', 'j_score'),
                               ('J_IDENTITY', 'j_identity'),
                               ('J_EVALUE', 'j_evalue'),
                               ('J_BTOP', 'j_btop'),
                               ('J_CIGAR', 'j_cigar'),
                               ('VDJ_SCORE', 'vdj_score'),
                               ('FWR1_IMGT', 'fwr1_imgt'),
                               ('FWR2_IMGT', 'fwr2_imgt'),
                               ('FWR3_IMGT', 'fwr3_imgt'),
                               ('FWR4_IMGT', 'fwr4_imgt'),
                               ('CDR1_IMGT', 'cdr1_imgt'),
                               ('CDR2_IMGT', 'cdr2_imgt'),
                               ('CDR3_IMGT', 'cdr3_imgt'),
                               ('FWR1_AA_IMGT', 'fwr1_aa_imgt'),
                               ('FWR2_AA_IMGT', 'fwr2_aa_imgt'),
                               ('FWR3_AA_IMGT', 'fwr3_aa_imgt'),
                               ('FWR4_AA_IMGT', 'fwr4_aa_imgt'),
                               ('CDR1_AA_IMGT', 'cdr1_aa_imgt'),
                               ('CDR2_AA_IMGT', 'cdr2_aa_imgt'),
                               ('CDR3_AA_IMGT', 'cdr3_aa_imgt'),
                               ('N1_LENGTH', 'n1_length'),
                               ('N2_LENGTH', 'n2_length'),
                               ('P3V_LENGTH', 'p3v_length'),
                               ('P5D_LENGTH', 'p5d_length'),
                               ('P3D_LENGTH', 'p3d_length'),
                               ('P5J_LENGTH', 'p5j_length'),
                               ('D_FRAME', 'd_frame'),
                               ('CDR3_IGBLAST', 'cdr3_igblast'),
                               ('CDR3_IGBLAST_AA', 'cdr3_igblast_aa'),
                               ('CONSCOUNT', 'conscount'),
                               ('DUPCOUNT', 'dupcount'),
                               ('UMICOUNT', 'umicount'),
                               ('CLONE', 'clone'),
                               ('CELL', 'cell')])

    # Mapping of Receptor attributes to Change-O column names
    _receptor_map = {v: k for k, v in _schema_map.items()}

    # All fields
    fields = list(_schema_map.keys())

    @staticmethod
    def toReceptor(field):
        """
        Returns a Receptor attribute name from a Change-O column name

        Arguments:
          field : Change-O column name.

        Returns:
          str: Receptor attribute name.
        """
        return ChangeoSchema._schema_map.get(field, field.lower())

    @staticmethod
    def fromReceptor(field):
        """
        Returns a Change-O column name from a Receptor attribute name

        Arguments:
          field : Receptor attribute name.

        Returns:
          str: Change-O column name.
        """
        return ChangeoSchema._receptor_map.get(field, field.upper())


class ChangeoSchemaAA(ChangeoSchema):
    """
    Change-O to Receptor amino acid mappings
    """
    # Standard fields
    required = ['SEQUENCE_ID',
                'SEQUENCE_AA_INPUT',
                'STOP',
                'INDELS',
                'LOCUS',
                'V_CALL',
                'SEQUENCE_AA_VDJ',
                'SEQUENCE_AA_IMGT',
                'V_SEQ_AA_START',
                'V_SEQ_AA_LENGTH',
                'V_GERM_AA_START_VDJ',
                'V_GERM_AA_LENGTH_VDJ',
                'V_GERM_AA_START_IMGT',
                'V_GERM_AA_LENGTH_IMGT',
                'GERMLINE_AA_IMGT']


class ReceptorData:
    """
    A class containing type conversion methods for Receptor data attributes

    Attributes:
      sequence_id (str): unique sequence identifier.

      rev_comp (bool): whether the alignment is relative to the reverse compliment of the input sequence.
      functional (bool): whether sample V(D)J sequence is predicted to be functional.
      in_frame (bool): whether junction region is in-frame.
      stop (bool): whether a stop codon is present in the V(D)J sequence.
      mutated_invariant (bool): whether the conserved amino acids are mutated in the V(D)J sequence.
      indels (bool): whether the V(D)J nucleotide sequence contains insertions and/or deletions.
      v_frameshift (bool): whether the V segment contains a frameshift

      sequence_input (Bio.Seq.Seq): input nucleotide sequence.
      sequence_vdj (Bio.Seq.Seq): Aligned V(D)J nucleotide sequence without IMGT-gaps.
      sequence_imgt (Bio.Seq.Seq): IMGT-gapped V(D)J nucleotide sequence.

      sequence_aa_input (Bio.Seq.Seq): input amino acid sequence.
      sequence_aa_vdj (Bio.Seq.Seq): Aligned V(D)J nucleotide sequence without IMGT-gaps.
      sequence_aa_imgt (Bio.Seq.Seq): IMGT-gapped V(D)J amino sequence.

      junction (Bio.Seq.Seq): ungapped junction region nucletide sequence.
      junction_aa (Bio.Seq.Seq): ungapped junction region amino acid sequence.
      junction_start (int): start positions of the junction in the input nucleotide sequence.
      junction_length (int): length of the junction in nucleotides.

      germline_vdj (Bio.Seq.Seq): full ungapped germline V(D)J nucleotide sequence.
      germline_vdj_d_mask (Bio.Seq.Seq): ungapped germline V(D)J nucleotides sequence with Ns masking the NP1-D-NP2 regions.
      germline_imgt (Bio.Seq.Seq): full IMGT-gapped germline V(D)J nucleotide sequence.
      germline_imgt_d_mask (Bio.Seq.Seq): IMGT-gapped germline V(D)J nucleotide sequence with ns masking the NP1-D-NP2 regions.

      germline_aa_vdj (Bio.Seq.Seq): full ungapped germline V(D)J amino acid sequence.
      germline_aa_imgt (Bio.Seq.Seq): full IMGT-gapped germline V(D)J amino acid sequence.

      v_call (str): V allele assignment(s).
      d_call (str): D allele assignment(s).
      j_call (str): J allele assignment(s).
      c_call (str): C region assignment.

      v_seq_start (int): position of the first V nucleotide in the input sequence (1-based).
      v_seq_length (int): number of V nucleotides in the input sequence.
      v_germ_start_imgt (int): position of the first V nucleotide in IMGT-gapped V germline sequence alignment (1-based).
      v_germ_length_imgt (int): length of the IMGT numbered germline V alignment.
      v_germ_start_vdj (int): position of the first nucleotide in ungapped V germline sequence alignment (1-based).
      v_germ_length_vdj (int): length of the ungapped germline V alignment.

      v_seq_aa_start (int): position of the first V amino acid in the amino acid input sequence (1-based).
      v_seq_aa_length (int): number of V amino acid in the amino acid input sequence.
      v_germ_aa_start_imgt (int): position of the first V amino acid in IMGT-gapped V germline amino acid alignment (1-based).
      v_germ_aa_length_imgt (int): length of the IMGT numbered germline V amino acid alignment.
      v_germ_aa_start_vdj (int): position of the first amino acid in ungapped V germline amino acid alignment (1-based).
      v_germ_aa_length_vdj (int): length of the ungapped germline V amino acid alignment.

      np1_start (int): position of the first untemplated nucleotide between the V and D segments in the input sequence (1-based).
      np1_length (int): number of untemplated nucleotides between the V and D segments.

      np1_aa_start (int): position of the first untemplated amino acid between the V and D segments in the input amino acid sequence (1-based).
      np1_aa_length (int): number of untemplated amino acids between the V and D segments.

      d_seq_start (int): position of the first D nucleotide in the input sequence (1-based).
      d_seq_length (int): number of D nucleotides in the input sequence.
      d_germ_start (int): position of the first nucleotide in D germline sequence alignment (1-based).
      d_germ_length (int): length of the germline D alignment.

      d_seq_aa_start (int): position of the first D amino acid in the input amino acidsequence (1-based).
      d_seq_aa_length (int): number of D amino acids in the input amino acid sequence.
      d_germ_aa_start (int): position of the first amino acid in D germline amino acid alignment (1-based).
      d_germ_aa_length (int): length of the germline D amino acid alignment.

      np2_start (int): position of the first untemplated nucleotide between the D and J segments in the input sequence (1-based).
      np2_length (int): number of untemplated nucleotides between the D and J segments.

      np2_aa_start (int): position of the first untemplated amino acid between the D and J segments in the input amino acid sequence (1-based).
      np2_aa_length (int): number of untemplated amino acid between the D and J segments.

      j_seq_start (int): position of the first J nucleotide in the input sequence (1-based).
      j_seq_length (int): number of J nucleotides in the input sequence.
      j_germ_start (int): position of the first nucleotide in J germline sequence alignment (1-based).
      j_germ_length (int): length of the germline J alignment.

      j_seq_aa_start (int): position of the first J amino acid in the input amino acidsequence (1-based).
      j_seq_aa_length (int): number of J amino acid in the input amino acidsequence.
      j_germ_aa_start (int): position of the first amino acid in J germline amino acid alignment (1-based).
      j_germ_aa_length (int): length of the germline J amino acid alignment.

      v_score (float): alignment score for the V.
      v_identity (float): alignment identity for the V.
      v_evalue (float): E-value for the alignment of the V.
      v_btop (str): BTOP for the alignment of the V.
      v_cigar (str): CIGAR for the alignment of the V.

      d_score (float): alignment score for the D.
      d_identity (float): alignment identity for the D.
      d_evalue (float): E-value for the alignment of the D.
      d_btop (str): BTOP for the alignment of the D.
      d_cigar (str): CIGAR for the alignment of the D.

      j_score (float): alignment score for the J.
      j_identity (float): alignment identity for the J.
      j_evalue (float): E-value for the alignment of the J.
      j_btop (str): BTOP for the alignment of the J.
      j_cigar (str): CIGAR for the alignment of the J.

      vdj_score (float): alignment score for the V(D)J.

      fwr1_imgt (Bio.Seq.Seq): IMGT-gapped FWR1 nucleotide sequence.
      fwr2_imgt (Bio.Seq.Seq): IMGT-gapped FWR2 nucleotide sequence.
      fwr3_imgt (Bio.Seq.Seq): IMGT-gapped FWR3 nucleotide sequence.
      fwr4_imgt (Bio.Seq.Seq): IMGT-gapped FWR4 nucleotide sequence.
      cdr1_imgt (Bio.Seq.Seq): IMGT-gapped CDR1 nucleotide sequence.
      cdr2_imgt (Bio.Seq.Seq): IMGT-gapped CDR2 nucleotide sequence.
      cdr3_imgt (Bio.Seq.Seq): IMGT-gapped CDR3 nucleotide sequence.
      cdr3_igblast (Bio.Seq.Seq): CDR3 nucleotide sequence assigned by IgBLAST.

      fwr1_aa_imgt (Bio.Seq.Seq): IMGT-gapped FWR1 amino acid sequence.
      fwr2_aa_imgt (Bio.Seq.Seq): IMGT-gapped FWR2 amino acid sequence.
      fwr3_aa_imgt (Bio.Seq.Seq): IMGT-gapped FWR3 amino acid sequence.
      fwr4_aa_imgt (Bio.Seq.Seq): IMGT-gapped FWR4 amino acid sequence.
      cdr1_aa_imgt (Bio.Seq.Seq): IMGT-gapped CDR1 amino acid sequence.
      cdr2_aa_imgt (Bio.Seq.Seq): IMGT-gapped CDR2 amino acid sequence.
      cdr3_aa_imgt (Bio.Seq.Seq): IMGT-gapped CDR3 amino acid sequence.
      cdr3_igblast_aa (Bio.Seq.Seq): CDR3 amino acid sequence assigned by IgBLAST.

      n1_length (int): M nucleotides 5' of the D segment.
      n2_length (int): nucleotides 3' of the D segment.
      p3v_length (int): palindromic nucleotides 3' of the V segment.
      p5d_length (int): palindromic nucleotides 5' of the D segment.
      p3d_length (int): palindromic nucleotides 3' of the D segment.
      p5j_length (int): palindromic nucleotides 5' of the J segment.
      d_frame (int): D segment reading frame.

      conscount (int): number of reads contributing to the UMI consensus sequence.
      dupcount (int): copy number of the sequence.
      umicount (int): number of UMIs representing the sequence.

      clone (str): clonal cluster identifier.
      cell (str): origin cell identifier.

      annotations (dict): dictionary containing all unknown fields.
    """
    #with resource_stream(__name__, 'data/receptor.yaml') as f:
    #    data = yaml.load(f, Loader=yaml.FullLoader)
    #
    # # Define type parsers
    # parsers = {k: v['type'] for k, v in data['receptor'].items()}
    #
    # # Define coordinate field sets
    # coordinates = {}
    # for k, v in data['receptor'].items():
    #     if 'coordinate' in v:
    #         position = {v['coordinate']['position']: k}
    #         group = coordinates.setdefault(v['coordinate']['group'], {})
    #         group.update(position)
    #
    # # Positional fields sets in the form {start: (length, end)}
    # self.start_fields = {x['start']: (x['length'], x['end']) for x in coordinates.values()}
    #
    # # Positional fields sets in the form {length: (start, end)}
    # self.length_fields = {x['length']: (x['start'], x['end']) for x in coordinates.values()}
    #
    # # Positional fields sets in the form {end: (start, length)}
    # self.end_fields = {x['end']: (x['start'], x['length']) for x in coordinates.values()}

    # Mapping of member variables to parsing functions
    parsers = {'sequence_id': 'identity',
               'rev_comp': 'logical',
               'functional': 'logical',
               'locus': 'identity',
               'in_frame': 'logical',
               'stop': 'logical',
               'mutated_invariant': 'logical',
               'indels': 'logical',
               'v_frameshift': 'logical',
               'sequence_input': 'nucleotide',
               'sequence_imgt': 'nucleotide',
               'sequence_vdj': 'nucleotide',
               'sequence_aa_input': 'aminoacid',
               'sequence_aa_imgt': 'aminoacid',
               'sequence_aa_vdj': 'aminoacid',
               'junction': 'nucleotide',
               'junction_aa': 'aminoacid',
               'junction_start': 'integer',
               'junction_length': 'integer',
               'germline_imgt': 'nucleotide',
               'germline_imgt_d_mask': 'nucleotide',
               'germline_vdj': 'nucleotide',
               'germline_vdj_d_mask': 'nucleotide',
               'germline_aa_imgt': 'aminoacid',
               'germline_aa_vdj': 'aminoacid',
               'v_call': 'identity',
               'd_call': 'identity',
               'j_call': 'identity',
               'c_call': 'identity',
               'v_seq_start': 'integer',
               'v_seq_length': 'integer',
               'v_germ_start_imgt': 'integer',
               'v_germ_length_imgt': 'integer',
               'v_germ_start_vdj': 'integer',
               'v_germ_length_vdj': 'integer',
               'v_seq_aa_start': 'integer',
               'v_seq_aa_length': 'integer',
               'v_germ_aa_start_imgt': 'integer',
               'v_germ_aa_length_imgt': 'integer',
               'v_germ_aa_start_vdj': 'integer',
               'v_germ_aa_length_vdj': 'integer',
               'np1_start': 'integer',
               'np1_length': 'integer',
               'np1_aa_start': 'integer',
               'np1_aa_length': 'integer',
               'd_seq_start': 'integer',
               'd_seq_length': 'integer',
               'd_germ_start': 'integer',
               'd_germ_length': 'integer',
               'd_seq_aa_start': 'integer',
               'd_seq_aa_length': 'integer',
               'd_germ_aa_start': 'integer',
               'd_germ_aa_length': 'integer',
               'np2_start': 'integer',
               'np2_length': 'integer',
               'np2_aa_start': 'integer',
               'np2_aa_length': 'integer',
               'j_seq_start': 'integer',
               'j_seq_length': 'integer',
               'j_germ_start': 'integer',
               'j_germ_length': 'integer',
               'j_seq_aa_start': 'integer',
               'j_seq_aa_length': 'integer',
               'j_germ_aa_start': 'integer',
               'j_germ_aa_length': 'integer',
               'v_score': 'double',
               'v_identity': 'double',
               'v_evalue': 'double',
               'v_btop': 'identity',
               'v_cigar': 'identity',
               'd_score': 'double',
               'd_identity': 'double',
               'd_evalue': 'double',
               'd_btop': 'identity',
               'd_cigar': 'identity',
               'j_score': 'double',
               'j_identity': 'double',
               'j_evalue': 'double',
               'j_btop': 'identity',
               'j_cigar': 'identity',
               'vdj_score': 'double',
               'fwr1_imgt': 'nucleotide',
               'fwr2_imgt': 'nucleotide',
               'fwr3_imgt': 'nucleotide',
               'fwr4_imgt': 'nucleotide',
               'cdr1_imgt': 'nucleotide',
               'cdr2_imgt': 'nucleotide',
               'cdr3_imgt': 'nucleotide',
               'fwr1_aa_imgt': 'aminoacid',
               'fwr2_aa_imgt': 'aminoacid',
               'fwr3_aa_imgt': 'aminoacid',
               'fwr4_aa_imgt': 'aminoacid',
               'cdr1_aa_imgt': 'aminoacid',
               'cdr2_aa_imgt': 'aminoacid',
               'cdr3_aa_imgt': 'aminoacid',
               'n1_length': 'integer',
               'n2_length': 'integer',
               'p3v_length': 'integer',
               'p5d_length': 'integer',
               'p3d_length': 'integer',
               'p5j_length': 'integer',
               'd_frame': 'integer',
               'cdr3_igblast': 'nucleotide',
               'cdr3_igblast_aa': 'aminoacid',
               'conscount': 'integer',
               'dupcount': 'integer',
               'umicount': 'integer',
               'clone': 'identity',
               'cell': 'identity'}

    # Positional fields sets in the form (start, length, end)
    _coordinate_map = [('v_seq_start', 'v_seq_length', 'v_seq_end'),
                       ('v_germ_start_imgt', 'v_germ_length_imgt', 'v_germ_end_imgt'),
                       ('v_germ_start_vdj', 'v_germ_length_vdj', 'v_germ_end_vdj'),
                       ('v_alignment_start', 'v_alignment_length', 'v_alignment_end'),
                       ('v_seq_aa_start', 'v_seq_aa_length', 'v_seq_aa_end'),
                       ('v_germ_aa_start_imgt', 'v_germ_aa_length_imgt', 'v_germ_aa_end_imgt'),
                       ('v_germ_aa_start_vdj', 'v_germ_aa_length_vdj', 'v_germ_aa_end_vdj'),
                       ('v_alignment_aa_start', 'v_alignment_aa_length', 'v_alignment_aa_end'),
                       ('d_seq_start', 'd_seq_length', 'd_seq_end'),
                       ('d_germ_start', 'd_germ_length', 'd_germ_end'),
                       ('d_seq_aa_start', 'd_seq_aa_length', 'd_seq_aa_end'),
                       ('d_germ_aa_start', 'd_germ_aa_length', 'd_germ_aa_end'),
                       ('j_seq_start', 'j_seq_length', 'j_seq_end'),
                       ('j_germ_start', 'j_germ_length', 'j_germ_end'),
                       ('j_seq_aa_start', 'j_seq_aa_length', 'j_seq_aa_end'),
                       ('j_germ_aa_start', 'j_germ_aa_length', 'j_germ_aa_end'),
                       ('junction_start', 'junction_length', 'junction_end'),
                       ('fwr1_start', 'fwr1_length', 'fwr1_end'),
                       ('fwr2_start', 'fwr2_length', 'fwr2_end'),
                       ('fwr3_start', 'fwr3_length', 'fwr3_end'),
                       ('fwr4_start', 'fwr4_length', 'fwr4_end'),
                       ('cdr1_start', 'cdr1_length', 'cdr1_end'),
                       ('cdr2_start', 'cdr2_length', 'cdr2_end'),
                       ('cdr3_start', 'cdr3_length', 'cdr3_end')]

    # Positional fields sets in the form {start: (length, end)}
    start_fields = {x[0]: (x[1], x[2]) for x in _coordinate_map}

    # Positional fields sets in the form {length: (start, end)}
    length_fields = {x[1]: (x[0], x[2]) for x in _coordinate_map}

    # Positional fields sets in the form {end: (start, length)}
    end_fields = {x[2]: (x[0], x[1]) for x in _coordinate_map}

    @staticmethod
    def identity(v, deparse=False):
        return v

    # Logical type conversion
    @staticmethod
    def logical(v, deparse=False):
        parse_map = {True: True, 'T': True, 'TRUE': True,
                     False: False, 'F': False, 'FALSE': False,
                     'NA': None, 'None': None, '': None}
        deparse_map = {False: 'F', True: 'T', None: ''}
        if not deparse:
            try:  return parse_map[v]
            except:  return None
        else:
            try:  return deparse_map[v]
            except:  return ''

    # Integer type conversion
    @staticmethod
    def integer(v, deparse=False):
        if not deparse:
            try:  return int(v)
            except:  return None
        else:
            return '' if v is None else str(v)

    # Float type conversion
    @staticmethod
    def double(v, deparse=False):
        if not deparse:
            try:  return float(v)
            except:  return None
        else:
            return '' if v is None else str(v)

    # Nucleotide sequence type conversion
    @staticmethod
    def nucleotide(v, deparse=False):
        if not deparse:
            try:
                #return '' if v in ('NA', 'None') else Seq(v, IUPAC.ambiguous_dna).upper()
                return '' if v in ('NA', 'None') else v.upper()
            except:
                return ''
        else:
            return '' if v in ('NA', 'None', None) else str(v)

    # Sequence type conversion
    @staticmethod
    def aminoacid(v, deparse=False):
        if not deparse:
            try:
                #return '' if v in ('NA', 'None') else Seq(v, IUPAC.extended_protein).upper()
                return '' if v in ('NA', 'None') else v.upper()
            except:
                return ''
        else:
            return '' if v in ('NA', 'None', None) else str(v)


class Receptor:
    """
    A class defining a V(D)J sequence and its annotations
    """
    # Mapping of derived properties to types
    _derived = {'v_seq_end': 'integer',
                'v_germ_end_vdj': 'integer',
                'v_germ_end_imgt': 'integer',
                'v_seq_aa_end': 'integer',
                'v_germ_aa_end_vdj': 'integer',
                'v_germ_aa_end_imgt': 'integer',
                'd_seq_end': 'integer',
                'd_germ_end': 'integer',
                'd_seq_aa_end': 'integer',
                'd_germ_aa_end': 'integer',
                'j_seq_end': 'integer',
                'j_germ_end': 'integer',
                'j_seq_aa_end': 'integer',
                'j_germ_aa_end': 'integer',
                'junction_end': 'integer'}

    def _junction_start(self):
        """
        Determine the position of the first junction nucleotide in the input sequence
        """
        try:
            x = self.v_germ_end_imgt - 310
            return self.v_seq_end - x if x >= 0 else None
        except TypeError:
            return None

    def __init__(self, data):
        """
        Initializer

        Arguments:
          data : dict of field/value data

        Returns:
          changeo.Receptor.Receptor
        """
        # Convert case of keys
        data = {k.lower(): v for k, v in data.items()}

        # Define known keys
        required_keys = ('sequence_id', )
        optional_keys = (x for x in ReceptorData.parsers if x not in required_keys)

        # Parse required fields
        try:
            for k in required_keys:
                f = getattr(ReceptorData, ReceptorData.parsers[k])
                setattr(self, k, f(data.pop(k)))
        except:
            printError('Input must contain valid %s values.' % ','.join(required_keys))

        # Parse optional known fields
        for k in optional_keys:
            f = getattr(ReceptorData, ReceptorData.parsers[k])
            setattr(self, k, f(data.pop(k, None)))

        # Derive junction_start if not provided
        if not hasattr(self, 'junction_start') or self.junction_start is None:
            setattr(self, 'junction_start', self._junction_start())

        # Add remaining elements as annotations dictionary
        self.annotations = data

    def setDict(self, data, parse=False):
        """
        Adds or updates multiple attributes and annotations

        Arguments:
          data : a dictionary of annotations to add or update.
          parse : if True pass values through string parsing functions for known fields.

        Returns:
          None : updates attribute values and the annotations attribute.
        """
        # Partition data
        attributes = {k.lower(): v for k, v in data.items() if k.lower() in ReceptorData.parsers}
        annotations = {k.lower(): v for k, v in data.items() if k.lower() not in attributes}

        # Update attributes
        for k, v in attributes.items():
            if parse:
                f = getattr(ReceptorData, ReceptorData.parsers[k])
                setattr(self, k, f(v))
            else:
                setattr(self, k, v)

        # Update annotations
        self.annotations.update(annotations)

    def setField(self, field, value, parse=False):
        """
        Set an attribute or annotation value

        Arguments:
          field : attribute name as a string
          value : value to assign
          parse : if True pass values through string parsing functions for known fields.

        Returns:
          None. Updates attribute or annotation.
        """
        field = field.lower()
        if field in ReceptorData.parsers and parse:
            f = getattr(ReceptorData, ReceptorData.parsers[field])
            setattr(self, field, f(value))
        elif field in ReceptorData.parsers:
            setattr(self, field, value)
        else:
            self.annotations[field] = value

    def getField(self, field):
        """
        Get an attribute or annotation value

        Arguments:
          field : attribute name as a string

        Returns:
          Value in the attribute. Returns None if the attribute cannot be found.
        """
        field = field.lower()

        if field in ReceptorData.parsers:
            return getattr(self, field)
        elif field in self.annotations:
            return self.annotations[field]
        else:
            return None

    def getSeq(self, field):
        """
        Get an attribute value converted to a Seq object

        Arguments:
          field : variable name as a string

        Returns:
          Bio.Seq.Seq : Value in the field as a Seq object
        """
        v = self.getField(field)

        if isinstance(v, Seq):
            return v
        elif isinstance(v, str):
            return Seq(v)
        else:
            return None

    def getAIRR(self, field, seq=False):
        """
        Get an attribute from an AIRR field name

        Arguments:
          field : AIRR column name as a string
          seq : if True return the attribute as a Seq object

        Returns:
          Value in the AIRR field. Returns None if the field cannot be found.
        """
        # Map to Receptor attribute
        field = AIRRSchema.toReceptor(field)

        if seq:
            return self.getSeq(field)
        else:
            return self.getField(field)

    def getChangeo(self, field, seq=False):
        """
        Get an attribute from a Change-O field name

        Arguments:
          field : Change-O column name as a string
          seq : if True return the attribute as a Seq object

        Returns:
          Value in the Change-O field. Returns None if the field cannot be found.
        """
        # Map to Receptor attribute
        field = ChangeoSchema.toReceptor(field)

        if seq:
            return self.getSeq(field)
        else:
            return self.getField(field)

    def toDict(self):
        """
        Convert the namespace to a dictionary

        Returns:
          dict : member fields with values converted to appropriate strings
        """
        d = {}
        n = self.__dict__
        # Parse attributes
        for k, v in n.items():
            if k == 'annotations':
                d.update(n['annotations'])
            else:
                f = getattr(ReceptorData, ReceptorData.parsers[k])
                d[k] = f(v, deparse=True)
        # Parse properties
        for k in Receptor._derived:
            f = getattr(ReceptorData, Receptor._derived[k])
            v = getattr(self, k)
            d[k] = f(v, deparse=True)
        return d

    def getAlleleCalls(self, calls, action='first'):
        """
        Get multiple allele calls

        Arguments:
          calls : iterable of calls to get; one or more of ('v','d','j')
          actions : One of ('first','set')

        Returns:
          list : List of requested calls in order
        """
        vdj = {'v': self.getVAllele(action),
               'd': self.getDAllele(action),
               'j': self.getJAllele(action)}

        return [vdj[k] for k in calls]

    def getGeneCalls(self, calls, action='first'):
        """
        Get multiple gene calls

        Arguments:
          calls : iterable of calls to get; one or more of ('v','d','j')
          actions : One of ('first','set')

        Returns:
          list : List of requested calls in order
        """
        vdj = {'v': self.getVGene(action),
               'd': self.getDGene(action),
               'j': self.getJGene(action)}

        return [vdj[k] for k in calls]

    def getFamilyCalls(self, calls, action='first'):
        """
        Get multiple family calls

        Arguments:
          calls : iterable of calls to get; one or more of ('v','d','j')
          actions : One of ('first','set')

        Returns:
          list : List of requested calls in order
        """
        vdj = {'v': self.getVFamily(action),
               'd': self.getDFamily(action),
               'j': self.getJFamily(action)}

        return [vdj[k] for k in calls]

    # TODO: this can't distinguish empty value ("") from missing field (no column)
    def getVAllele(self, action='first', field=None):
        """
        V segment allele getter

        Arguments:
          actions : One of 'first', 'set' or list'
          field : attribute or annotation name containing the V call. Use v_call attribute if None.

        Returns:
          str : String of the allele when action is 'first';
          tuple : Tuple of allele calls for 'set' or 'list' actions.
        """
        x = self.v_call if field is None else self.getField(field)
        return getAllele(x, action=action)

    def getDAllele(self, action='first', field=None):
        """
        D segment allele getter

        Arguments:
          actions : One of 'first', 'set' or 'list'
          field : attribute or annotation name containing the D call. Use d_call attribute if None.

        Returns:
          str : String of the allele when action is 'first';
          tuple : Tuple of allele calls for 'set' or 'list' actions.
        """
        x = self.d_call if field is None else self.getField(field)
        return getAllele(x, action=action)

    def getJAllele(self, action='first', field=None):
        """
        J segment allele getter

        Arguments:
          actions : One of 'first', 'set' or 'list'
          field : attribute or annotation name containing the J call. Use j_call attribute if None.

        Returns:
          str : String of the allele when action is 'first';
          tuple : Tuple of allele calls for 'set' or 'list' actions.
        """
        x = self.j_call if field is None else self.getField(field)
        return getAllele(x, action=action)

    def getVGene(self, action='first', field=None):
        """
        V segment gene getter

        Arguments:
          actions : One of 'first', 'set' or list'
          field : attribute or annotation name containing the V call. Use v_call attribute if None.

        Returns:
          str : String of the allele when action is 'first';
          tuple : Tuple of allele calls for 'set' or 'list' actions.
        """
        x = self.v_call if field is None else self.getField(field)
        return getGene(x, action=action)

    def getDGene(self, action='first', field=None):
        """
        D segment gene getter

        Arguments:
          actions : One of 'first', 'set' or list'
          field : attribute or annotation name containing the D call. Use d_call attribute if None.

        Returns:
          str : String of the allele when action is 'first';
          tuple : Tuple of allele calls for 'set' or 'list' actions.
        """
        x = self.d_call if field is None else self.getField(field)
        return getGene(x, action=action)

    def getJGene(self, action='first', field=None):
        """
        J segment gene getter

        Arguments:
          actions : One of 'first', 'set' or list'
          field : attribute or annotation name containing the J call. Use j_call attribute if None.

        Returns:
          str : String of the allele when action is 'first';
          tuple : Tuple of allele calls for 'set' or 'list' actions.
        """
        x = self.j_call if field is None else self.getField(field)
        return getGene(x, action=action)

    def getVFamily(self, action='first', field=None):
        """
        V segment family getter

        Arguments:
          actions : One of 'first', 'set' or list'
          field : attribute or annotation name containing the V call. Use v_call attribute if None.

        Returns:
          str : String of the allele when action is 'first';
          tuple : Tuple of allele calls for 'set' or 'list' actions.
        """
        x = self.v_call if field is None else self.getField(field)
        return getFamily(x, action=action)

    def getDFamily(self, action='first', field=None):
        """
        D segment family getter

        Arguments:
          actions : One of 'first', 'set' or list'
          field : attribute or annotation name containing the D call. Use d_call attribute if None.

        Returns:
          str : String of the allele when action is 'first';
          tuple : Tuple of allele calls for 'set' or 'list' actions.
        """
        x = self.d_call if field is None else self.getField(field)
        return getFamily(x, action=action)

    def getJFamily(self, action='first', field=None):
        """
        J segment family getter

        Arguments:
          actions : One of 'first', 'set' or list'
          field : attribute or annotation name containing the J call. Use j_call attribute if None.

        Returns:
          str : String of the allele when action is 'first';
          tuple : Tuple of allele calls for 'set' or 'list' actions.
        """
        x = self.j_call if field is None else self.getField(field)
        return getFamily(x, action=action)

    def getAlleleNumbers(self, calls, action='first'):
        """
        Get multiple allele numeric identifiers

        Arguments:
          calls : iterable of calls to get; one or more of ('v','d','j')
          actions : One of ('first','set')

        Returns:
          list : List of requested calls in order
        """
        vdj = {'v': self.getVAlleleNumber(action),
               'd': self.getDAlleleNumber(action),
               'j': self.getJAlleleNumber(action)}

        return [vdj[k] for k in calls]

    def getVAlleleNumber(self, action='first', field=None):
        """
        V segment allele number getter

        Arguments:
          actions : One of 'first', 'set' or list'
          field : attribute or annotation name containing the V call. Use v_call attribute if None.

        Returns:
          str : String of the allele when action is 'first';
          tuple : Tuple of allele numbers for 'set' or 'list' actions.
        """
        x = self.v_call if field is None else self.getField(field)
        return getAlleleNumber(x, action=action)

    def getDAlleleNumber(self, action='first', field=None):
        """
        D segment allele number getter

        Arguments:
          actions : One of 'first', 'set' or list'
          field : attribute or annotation name containing the D call. Use d_call attribute if None.

        Returns:
          str : String of the allele when action is 'first';
          tuple : Tuple of allele numbers for 'set' or 'list' actions.
        """
        x = self.d_call if field is None else self.getField(field)
        return getAlleleNumber(x, action=action)

    def getJAlleleNumber(self, action='first', field=None):
        """
        J segment allele number getter

        Arguments:
          actions : One of 'first', 'set' or list'
          field : attribute or annotation name containing the J call. Use j_call attribute if None.

        Returns:
          str : String of the allele when action is 'first';
          tuple : Tuple of allele numbers for 'set' or 'list' actions.
        """
        x = self.j_call if field is None else self.getField(field)
        return getAlleleNumber(x, action=action)

    @property
    def v_seq_end(self):
        """
        Position of the last V nucleotide in the input sequence
        """
        try:  return self.v_seq_start + self.v_seq_length - 1
        except TypeError:  return None

    @property
    def v_germ_end_imgt(self):
        """
        Position of the last nucleotide in the IMGT-gapped V germline sequence alignment
        """
        try:  return self.v_germ_start_imgt + self.v_germ_length_imgt - 1
        except TypeError:  return None

    @property
    def v_germ_end_vdj(self):
        """
        Position of the last nucleotide in the ungapped V germline sequence alignment
        """
        try:  return self.v_germ_start_vdj + self.v_germ_length_vdj - 1
        except TypeError:  return None

    @property
    def v_seq_aa_end(self):
        """
        Position of the last V nucleotide in the input sequence
        """
        try:  return self.v_seq_aa_start + self.v_seq_aa_length - 1
        except TypeError:  return None

    @property
    def v_germ_aa_end_imgt(self):
        """
        Position of the last nucleotide in the IMGT-gapped V germline sequence alignment
        """
        try:  return self.v_germ_aa_start_imgt + self.v_germ_aa_length_imgt - 1
        except TypeError:  return None

    @property
    def v_germ_aa_end_vdj(self):
        """
        Position of the last nucleotide in the ungapped V germline sequence alignment
        """
        try:  return self.v_germ_aa_start_vdj + self.v_germ_aa_length_vdj - 1
        except TypeError:  return None

    @property
    def d_seq_end(self):
        """
        Position of the last D nucleotide in the input sequence
        """
        try:  return self.d_seq_start + self.d_seq_length - 1
        except TypeError:  return None

    @property
    def d_germ_end(self):
        """
        Position of the last nucleotide in the D germline sequence alignment
        """
        try:  return self.d_germ_start + self.d_germ_length - 1
        except TypeError:  return None

    @property
    def d_seq_aa_end(self):
        """
        Position of the last D amino acid in the input amino acid sequence
        """
        try:  return self.d_seq_aa_start + self.d_seq_aa_length - 1
        except TypeError:  return None

    @property
    def d_germ_aa_end(self):
        """
        Position of the last amino acid in the D germline amino acid alignment
        """
        try:  return self.d_germ_aa_start + self.d_germ_aa_length - 1
        except TypeError:  return None

    @property
    def j_seq_end(self):
        """
        Position of the last J nucleotide in the input sequence
        """
        try:  return self.j_seq_start + self.j_seq_length - 1
        except TypeError:  return None

    @property
    def j_germ_end(self):
        """
        Position of the last nucleotide in the J germline sequence alignment
        """
        try:  return self.j_germ_start + self.j_germ_length - 1
        except TypeError:  return None

    @property
    def j_seq_aa_end(self):
        """
        Position of the last J amino acid in the input amino sequence
        """
        try:  return self.j_seq_aa_start + self.j_seq_aa_length - 1
        except TypeError:  return None

    @property
    def j_germ_aa_end(self):
        """
        Position of the last amino acid in the J germline amino acid alignment
        """
        try:  return self.j_germ_aa_start + self.j_germ_aa_length - 1
        except TypeError:  return None

    @property
    def junction_end(self):
        """
        Position of the last junction nucleotide in the input sequence
        """
        try:
            gaps = self.junction.count('.')
            return self.junction_start + self.junction_length - gaps - 1
        except TypeError:
            return None



"""
Gene annotations
"""

# Info
__author__ = 'Jason Anthony Vander Heiden'

# Imports
import re
from collections import OrderedDict

# Presto and changeo imports
from changeo.Defaults import v_attr, d_attr, j_attr, seq_attr

# Ig and TCR Regular expressions
allele_number_regex = re.compile(r'(?<=\*)([\.\w]+)')
allele_regex = re.compile(r'((IG[HLK]|TR[ABGD])([VDJ][A-R0-9]+[-/\w]*[-\*][\.\w]+))')
gene_regex = re.compile(r'((IG[HLK]|TR[ABGD])([VDJ][A-R0-9]+[-/\w]*))')
family_regex = re.compile(r'((IG[HLK]|TR[ABGD])([VDJ][A-R0-9]+))')
locus_regex = re.compile(r'(IG[HLK]|TR[ABGD])')

v_allele_regex = re.compile(r'((IG[HLK]|TR[ABGD])V[A-R0-9]+[-/\w]*[-\*][\.\w]+)')
d_allele_regex = re.compile(r'((IG[HLK]|TR[ABGD])D[A-R0-9]+[-/\w]*[-\*][\.\w]+)')
j_allele_regex = re.compile(r'((IG[HLK]|TR[ABGD])J[A-R0-9]+[-/\w]*[-\*][\.\w]+)')
c_allele_regex = re.compile(r'((IG[HLK]|TR[ABGD])([DMAGEC][P0-9]?[A-Z]?)[\*][\.\w]+)')
c_gene_regex = re.compile(r'((IG[HLK]|TR[ABGD])([DMAGEC][P0-9]?[A-Z]?))')


def parseGeneCall(gene, regex, action='first'):
    """
    Extract alleles from strings

    Arguments:
      gene (str): string with gene calls
      regex (re.Pattern): compiled regular expression for allele match
      action (str): action to perform for multiple alleles;
               one of ('first', 'set', 'list').
    Returns:
      str: String of the allele when action is 'first';
      tuple: Tuple of allele calls for 'set' or 'list' actions.
    """
    try:
        match = [x.group(0) for x in regex.finditer(gene)]
    except:
        match = None

    if action == 'first':
        return match[0] if match else None
    elif action == 'set':
        return tuple(sorted(set(match))) if match else None
    elif action == 'list':
        return tuple(sorted(match)) if match else None
    else:
        return None


def getAllele(gene, action='first'):
    """
    Extract allele from gene call string

    Arguments:
      gene (str): string with gene calls
      action (str): action to perform for multiple alleles;
                    one of ('first', 'set', 'list').

    Returns:
      str: String of the first allele calls when action is 'first'.
      tuple: Tuple of allele calls for 'set' or 'list' actions.
    """
    return parseGeneCall(gene, allele_regex, action=action)


def getGene(gene, action='first'):
    """
    Extract gene from gene call string

    Arguments:
      gene (str): string with gene calls
      action (str): action to perform for multiple alleles;
                    one of ('first', 'set', 'list').

    Returns:
      str: String of the first gene call when action is 'first'.
      tuple: Tuple of gene calls for 'set' or 'list' actions.
    """
    return parseGeneCall(gene, gene_regex, action=action)


def getFamily(gene, action='first'):
    """
    Extract family from gene call string

    Arguments:
      gene (str): string with gene calls
      action (str): action to perform for multiple alleles;
                    one of ('first', 'set', 'list').

    Returns:
      str: String of the first family call when action is 'first'.
      tuple: Tuple of allele calls for 'set' or 'list' actions.
    """
    return parseGeneCall(gene, family_regex, action=action)


def getLocus(gene, action='first'):
    """
    Extract locus from gene call string

    Arguments:
      gene (str): string with gene calls
      action (str): action to perform for multiple alleles;
                    one of ('first', 'set', 'list').

    Returns:
      str: String of the first locus call when action is 'first'.
      tuple: Tuple of locus calls for 'set' or 'list' actions.
    """
    return parseGeneCall(gene, locus_regex, action=action)


def getAlleleNumber(gene, action='first'):
    """
    Extract allele number from gene call string

    Arguments:
      gene (str): string with gene calls
      action (str): action to perform for multiple alleles;
                    one of ('first', 'set', 'list').

    Returns:
      str: String of the first allele number call when action is 'first'.
      tuple: Tuple of allele numbers for 'set' or 'list' actions.
    """
    return parseGeneCall(gene, allele_number_regex, action=action)


def getVAllele(gene, action='first'):
    """
    Extract V allele gene from gene call string

    Arguments:
      gene (str): string with V gene calls
      action (str): action to perform for multiple alleles;
                    one of ('first', 'set', 'list').

    Returns:
      str: String of the first V allele call when action is 'first'.
      tuple: Tuple of V allele calls for 'set' or 'list' actions.
    """
    return parseGeneCall(gene, v_allele_regex, action=action)


def getDAllele(gene, action='first'):
    """
    Extract D allele gene from gene call string

    Arguments:
      gene (str): string with D gene calls
      action (str): action to perform for multiple alleles;
                    one of ('first', 'set', 'list').

    Returns:
      str: String of the first D allele call when action is 'first'.
      tuple: Tuple of D allele calls for 'set' or 'list' actions.
    """
    return parseGeneCall(gene, d_allele_regex, action=action)


def getJAllele(gene, action='first'):
    """
    Extract J allele gene from gene call string

    Arguments:
      gene (str): string with J gene calls
      action (str): action to perform for multiple alleles;
                    one of ('first', 'set', 'list').

    Returns:
      str: String of the first J allele call when action is 'first'.
      tuple: Tuple of J allele calls for 'set' or 'list' actions.
    """
    return parseGeneCall(gene, j_allele_regex, action=action)


def getCAllele(gene, action='first'):
    """
    Extract C-region allele gene call string

    Arguments:
      gene (str): string with C-region gene calls
      action (str): action to perform for multiple alleles;
                    one of ('first', 'set', 'list').

    Returns:
      str: String of the first C-region allele call when action is 'first'.
      tuple: Tuple of allele calls for 'set' or 'list' actions.
    """
    return parseGeneCall(gene, c_allele_regex, action=action)


def getCGene(gene, action='first'):
    """
    Extract C-region gene from gene call string

    Arguments:
      gene (str): string with C-region gene calls
      action (str): action to perform for multiple alleles;
                    one of ('first', 'set', 'list').

    Returns:
      str: String of the first C-region gene call when action is 'first'.
      tuple: Tuple of gene calls for 'set' or 'list' actions.
    """
    return parseGeneCall(gene, c_gene_regex, action=action)


# TODO: this is not generalized for non-IMGT gapped sequences!
def getVGermline(receptor, references, v_field=v_attr, amino_acid=False):
    """
    Extract V allele and germline sequence

    Arguments:
      receptor (changeo.Receptor.Receptor): Receptor object
      references (dict): dictionary of germline sequences
      v_field (str): Receptor attribute containing the V allele assignment
      amino_acid (bool): if True then use the amino acid positional fields, otherwise use the nucleotide fields.

    Returns:
      tuple: V allele name, V segment germline sequence
    """
    # Extract V allele call
    vgene = receptor.getVAllele(action='first', field=v_field)

    # Get germline start and length
    if not amino_acid:
        pad_char = 'N'
        try:  vstart = int(receptor.v_germ_start_imgt) - 1
        except (TypeError, ValueError):  vstart = 0
        try:  vlen = int(receptor.v_germ_length_imgt)
        except (TypeError, ValueError):  vlen = 0
    else:
        pad_char = 'X'
        try:  vstart = int(receptor.v_germ_aa_start_imgt) - 1
        except (TypeError, ValueError, AttributeError):  vstart = 0
        try:  vlen = int(receptor.v_germ_aa_length_imgt)
        except (TypeError, ValueError, AttributeError):  vlen = 0

    # Build V segment germline sequence
    if vgene is None:
        germ_vseq = pad_char * vlen
    elif vgene in references:
        vseq = references[vgene]
        vpad = vlen - len(vseq[vstart:])
        if vpad < 0: vpad = 0
        germ_vseq = vseq[vstart:(vstart + vlen)] + (pad_char * vpad)
    else:
        germ_vseq = None

    return vgene, germ_vseq


def getDGermline(receptor, references, d_field=d_attr, amino_acid=False):
    """
    Extract D allele and germline sequence

    Arguments:
      receptor (changeo.Receptor.Receptor): Receptor object
      references (dict): dictionary of germline sequences
      d_field (str): Receptor attribute containing the D allele assignment
      amino_acid (bool): if True then use the amino acid positional fields, otherwise use the nucleotide fields.


    Returns:
      tuple: D allele name, D segment germline sequence
    """
    # Extract D allele call
    dgene = receptor.getDAllele(action='first', field=d_field)

    # Get germline start and length
    if not amino_acid:
        try:  dstart = int(receptor.d_germ_start) - 1
        except (TypeError, ValueError):  dstart = 0
        try:  dlen = int(receptor.d_germ_length)
        except (TypeError, ValueError):  dlen = 0
    else:
        try:  dstart = int(receptor.d_germ_aa_start) - 1
        except (TypeError, ValueError, AttributeError):  dstart = 0
        try:  dlen = int(receptor.d_germ_aa_length)
        except (TypeError, ValueError, AttributeError):  dlen = 0

    # Build D segment germline sequence
    if dgene is None:
        germ_dseq = ''
    elif dgene in references:
        # Define D germline sequence
        dseq = references[dgene]
        germ_dseq = dseq[dstart:(dstart + dlen)]
    else:
        germ_dseq = None

    return dgene, germ_dseq


def getJGermline(receptor, references, j_field=j_attr, amino_acid=False):
    """
    Extract J allele and germline sequence

    Arguments:
      receptor (changeo.Receptor.Receptor): Receptor object
      references (dict): dictionary of germline sequences
      j_field (str): Receptor attribute containing the J allele assignment
      amino_acid (bool): if True then use the amino acid positional fields, otherwise use the nucleotide fields.

    Returns:
      tuple: J allele name, J segment germline sequence
    """
    # Extract J allele call
    jgene = receptor.getJAllele(action='first', field=j_field)

    # Get germline start and length
    if not amino_acid:
        pad_char = 'N'
        try:  jstart = int(receptor.j_germ_start) - 1
        except (TypeError, ValueError):  jstart = 0
        try:  jlen = int(receptor.j_germ_length)
        except (TypeError, ValueError):  jlen = 0
    else:
        pad_char = 'X'
        try:  jstart = int(receptor.j_germ_aa_start) - 1
        except (TypeError, ValueError, AttributeError):  jstart = 0
        try:  jlen = int(receptor.j_germ_aa_length)
        except (TypeError, ValueError, AttributeError):  jlen = 0

    # Build J segment germline sequence
    if jgene is None:
        germ_jseq = pad_char * jlen
    elif jgene in references:
        jseq = references[jgene]
        jpad = jlen - len(jseq[jstart:])
        if jpad < 0: jpad = 0
        germ_jseq = jseq[jstart:(jstart + jlen)] + (pad_char * jpad)
    else:
        germ_jseq = None

    return jgene, germ_jseq


def stitchVDJ(receptor, v_seq, d_seq, j_seq, amino_acid=False):
    """
    Assemble full length germline sequence

    Arguments:
      receptor (changeo.Receptor.Receptor): Receptor object
      v_seq (str): V segment sequence as a string
      d_seq (str): D segment sequence as a string
      j_seq (str): J segment sequence as a string
      amino_acid (bool): if True use X for N/P regions and amino acid positional fields,
                         otherwise use N and nucleotide fields.

    Returns:
      str: full germline sequence
    """
    # Get N/P lengths
    if not amino_acid:
        np_char = 'N'
        try:  np1_len = int(receptor.np1_length)
        except (TypeError, ValueError):  np1_len = 0
        try:  np2_len = int(receptor.np2_length)
        except (TypeError, ValueError):  np2_len = 0
    else:
        np_char = 'X'
        try:  np1_len = int(receptor.np1_aa_length)
        except (TypeError, ValueError, AttributeError):  np1_len = 0
        try:  np2_len = int(receptor.np2_aa_length)
        except (TypeError, ValueError, AttributeError):  np2_len = 0

    # Assemble pieces starting with V segment
    sequence = v_seq
    sequence += np_char * np1_len
    sequence += d_seq
    sequence += np_char * np2_len
    sequence += j_seq

    return sequence


def stitchRegions(receptor, v_seq, d_seq, j_seq, amino_acid=False):
    """
    Assemble full length region encoding

    Arguments:
      receptor (changeo.Receptor.Receptor): Receptor object
      v_seq (str): V segment germline sequence as a string
      d_seq (str): D segment germline sequence as a string
      j_seq (str): J segment germline sequence as a string
      amino_acid (bool): if True use amino acid positional fields, otherwise use nucleotide fields.

    Returns:
      str: string defining germline regions
    """
    # Set mode for region definitions
    full_junction = True if getattr(receptor, 'n1_length', None) is not None else False

    # Assemble pieces starting with V segment
    regions = 'V' * len(v_seq)

    # NP nucleotide additions after V
    if amino_acid:
        # PNP nucleotide additions after V
        try:  np1_len = int(receptor.np1_aa_length)
        except (TypeError, ValueError, AttributeError):  np1_len = 0
        regions += 'N' * np1_len
    elif not full_junction:
        # PNP nucleotide additions after V
        try:  np1_len = int(receptor.np1_length)
        except (TypeError, ValueError):  np1_len = 0
        regions += 'N' * np1_len
    else:
        # P nucleotide additions before N1
        try:  p3v_len = int(receptor.p3v_length)
        except (TypeError, ValueError):  p3v_len = 0
        # N1 nucleotide additions
        try:  n1_len = int(receptor.n1_length)
        except (TypeError, ValueError):  n1_len = 0
        # P nucleotide additions before D
        try:  p5d_len = int(receptor.p5d_length)
        except (TypeError, ValueError):  p5d_len = 0

        # Update regions
        regions += 'P' * p3v_len
        regions += 'N' * n1_len
        regions += 'P' * p5d_len

    # Add D segment
    regions += 'D' * len(d_seq)

    # NP nucleotide additions before J
    if amino_acid:
        # NP nucleotide additions
        try:  np2_len = int(receptor.np2_aa_length)
        except (TypeError, ValueError, AttributeError):  np2_len = 0
        regions += 'N' * np2_len
    elif not full_junction:
        # NP nucleotide additions
        try:  np2_len = int(receptor.np2_length)
        except (TypeError, ValueError):  np2_len = 0
        regions += 'N' * np2_len
    else:
        # P nucleotide additions after D
        try: p3d_len = int(receptor.p3d_length)
        except (TypeError, ValueError): p3d_len = 0
        # N2 nucleotide additions
        try:  n2_len = int(receptor.n2_length)
        except (TypeError, ValueError): n2_len = 0
        # P nucleotide additions before J
        try:  p5j_len = int(receptor.p5j_length)
        except (TypeError, ValueError):  p5j_len = 0

        # Update regions
        regions += 'P' * p3d_len
        regions += 'N' * n2_len
        regions += 'P' * p5j_len

    # Add J segment
    regions += 'J' * len(j_seq)

    return regions


# TODO: Should do 'first' method for ambiguous V/J groups. And explicit allele extraction.
def buildGermline(receptor, references, seq_field=seq_attr, v_field=v_attr,
                  d_field=d_attr, j_field=j_attr, amino_acid=False):
    """
    Join gapped germline sequences aligned with sample sequences

    Arguments:
      receptor (changeo.Receptor.Receptor): Receptor object.
      references (dict): dictionary of IMGT gapped germline sequences.
      seq_field (str): Receptor attribute in which to look for sequence.
      v_field (str): Receptor attribute in which to look for V call.
      d_field (str): Receptor attribute in which to look for V call.
      j_field (str): Receptor attribute in which to look for V call.
      amino_acid (bool): if True then use the amino acid positional fields, otherwise use the nucleotide fields.

    Returns:
      tuple: log dictionary, dictionary of {germline_type: germline_sequence}, dictionary of {segment: gene call}
    """
    # Return objects
    log = OrderedDict()
    germlines = {'full': '', 'dmask': '', 'vonly': '', 'regions': ''}

    # Build V segment germline sequence
    vgene, germ_vseq = getVGermline(receptor, references, v_field=v_field, amino_acid=amino_acid)
    log['V_CALL'] = vgene
    if germ_vseq is None:
        log['ERROR'] = 'Allele %s is not in the provided germline database.' % vgene
        return log, None, None

    # Build D segment germline sequence
    dgene, germ_dseq = getDGermline(receptor, references, d_field=d_field, amino_acid=amino_acid)
    log['D_CALL'] = dgene
    if germ_dseq is None:
        log['ERROR'] = 'Allele %s is not in the provided germline database.' % dgene
        return log, None, None

    # Build J segment germline sequence
    jgene, germ_jseq = getJGermline(receptor, references, j_field=j_field, amino_acid=amino_acid)
    log['J_CALL'] = jgene
    if germ_jseq is None:
        log['ERROR'] = 'Allele %s is not in the provided germline database.' % jgene
        return log, None, None

    # Stitch complete germlines
    germ_seq = stitchVDJ(receptor, germ_vseq, germ_dseq, germ_jseq, amino_acid=amino_acid)
    regions = stitchRegions(receptor, germ_vseq, germ_dseq, germ_jseq, amino_acid=amino_acid)

    # Update log
    log['SEQUENCE'] = receptor.getField(seq_field)
    log['GERMLINE'] = germ_seq
    log['REGIONS'] = regions

    # Check that input and germline sequence match
    if len(receptor.getField(seq_field)) == 0:
        log['ERROR'] = 'Sequence is missing from the %s field' % seq_field
        return log, None, None

    len_check = len(germ_seq) - len(receptor.getField(seq_field))
    if len_check != 0:
        log['ERROR'] = 'Germline sequence differs in length from input sequence by %i characters.' % abs(len_check)
        return log, None, None

    # Define return germlines object
    pad_char = 'X' if amino_acid else 'N'
    germ_dmask = germ_seq[:len(germ_vseq)] + \
                 pad_char * (len(germ_seq) - len(germ_vseq) - len(germ_jseq)) + \
                 germ_seq[-len(germ_jseq):]
    germlines = {'full': germ_seq, 'dmask': germ_dmask, 'vonly': germ_vseq, 'regions': regions}
    for k, v in germlines.items():  germlines[k] = v.upper()

    # Define return genes object
    genes = {'v': log['V_CALL'],
             'd': log['D_CALL'],
             'j': log['J_CALL']}

    return log, germlines, genes


def buildClonalGermline(receptors, references, seq_field=seq_attr, v_field=v_attr,
                        d_field=d_attr, j_field=j_attr, amino_acid=False):
    """
    Determine consensus clone sequence and create germline for clone

    Arguments:
      receptors (changeo.Receptor.Receptor): list of Receptor objects
      references (dict): dictionary of IMGT gapped germline sequences
      seq_field (str): Receptor attribute in which to look for sequence
      v_field (str): Receptor attributein which to look for V call
      d_field (str): Receptor attributein which to look for D call
      j_field (str): Receptor attributein which to look for J call
      amino_acid (bool): if True then use the amino acid positional fields, otherwise use the nucleotide fields.


    Returns:
      tuple: log dictionary, dictionary of {germline_type: germline_sequence},
              dictionary of consensus {segment: gene call}
    """
    # Log
    log = OrderedDict()

    # Create dictionaries to count observed V/J calls
    v_dict = OrderedDict()
    j_dict = OrderedDict()

    # Amino acid settings
    pad_char = 'X' if amino_acid else 'N'

    # Find longest sequence in clone
    max_length = 0
    for rec in receptors:
        v = rec.getVAllele(action='first', field=v_field)
        v_dict[v] = v_dict.get(v, 0) + 1
        j = rec.getJAllele(action='first', field=j_field)
        j_dict[j] = j_dict.get(j, 0) + 1
        seq_len = len(rec.getField(seq_field))
        if seq_len > max_length:
            max_length = seq_len

    # Consensus V and J having most observations
    v_cons = [k for k in list(v_dict.keys()) if v_dict[k] == max(v_dict.values())]
    j_cons = [k for k in list(j_dict.keys()) if j_dict[k] == max(j_dict.values())]

    # Consensus sequence(s) with consensus V/J calls and longest sequence
    cons = [x for x in receptors if x.getVAllele(action='first', field=v_field) in v_cons and \
                                    x.getJAllele(action='first', field=j_field) in j_cons and \
                                    len(x.getField(seq_field)) == max_length]
    # Consensus sequence(s) with consensus V/J calls but not the longest sequence
    if not cons:
        cons = [x for x in receptors if x.getVAllele(action='first', field=v_field) in v_cons and \
                                        x.getJAllele(action='first', field=j_field) in j_cons]

    # Return without germline if no sequence has both consensus V and J call
    if not cons:
        log['V_CALL'] = ','.join(v_cons)
        log['J_CALL'] = ','.join(j_cons)
        log['ERROR'] = 'No sequence found with both consensus V and J calls.'
        return log, None, None

    # Select consensus Receptor, resolving ties by alphabetical ordering of sequence id.
    cons = sorted(cons, key=lambda x: x.sequence_id)[0]

    # Pad end of consensus sequence with gaps to make it the max length
    gap_length = max_length - len(cons.getField(seq_field))
    if gap_length > 0:
        if amino_acid:  cons.j_germ_aa_length = int(cons.j_germ_aa_length or 0) + gap_length
        else:  cons.j_germ_length = int(cons.j_germ_length or 0) + gap_length
        cons.setField(seq_field, cons.getField(seq_field) + (pad_char * gap_length))

    # Update lengths padded to longest sequence in clone
    for rec in receptors:
        x = max_length - len(rec.getField(seq_field))
        if amino_acid:  rec.j_germ_aa_length = int(rec.j_germ_aa_length or 0) + x
        else:  rec.j_germ_length = int(rec.j_germ_length or 0) + x
        rec.setField(seq_field, rec.getField(seq_field) + (pad_char * x))

    # Stitch consensus germline
    cons_log, germlines, genes = buildGermline(cons, references, seq_field=seq_field, v_field=v_field,
                                               d_field=d_field, j_field=j_field, amino_acid=amino_acid)

    # Update log
    log['CONSENSUS'] = cons.sequence_id
    log.update(cons_log)

    # Return log
    return log, germlines, genes



"""
Unit tests for ParseDb
"""

# Info
__author__ = 'Jason Anthony Vander Heiden'
from changeo import __version__, __date__

# Imports
import os
import sys
import time
import unittest

# Paths
test_path = os.path.dirname(os.path.realpath(__file__))
data_path = os.path.join(test_path, 'data')

# Import script
sys.path.append(os.path.join(test_path, os.pardir, 'bin'))
import ParseDb


class Test_ParseDb(unittest.TestCase):
    def setUp(self):
        print('-> %s()' % self._testMethodName)
        self.start = time.time()

    def tearDown(self):
        t = time.time() - self.start
        print("<- %s() %.3f" % (self._testMethodName, t))


if __name__ == '__main__':
    unittest.main()


"""
Multiprocessing
"""

# Info
__author__ = 'Jason Anthony Vander Heiden'

# Imports
import os
import sys
from collections import OrderedDict
from time import time

# Presto and changeo imports
from presto.IO import printProgress, printLog, printError, printWarning
from changeo.Defaults import default_out_args
from changeo.IO import countDbFile, getOutputHandle, AIRRReader, AIRRWriter
from changeo.Receptor import Receptor


class DbData:
    """
    A class defining data objects for worker processes

    Attributes:
      id : result identifier
      data : list of data records
      valid : True if preprocessing was successfull and data should be processed
    """
    # Instantiation
    def __init__(self, key, records):
        self.id = key
        self.data = records
        self.valid = (key is not None and records is not None)

    # Boolean evaluation
    def __bool__(self):
        return self.valid

    # Length evaluation
    def __len__(self):
        if isinstance(self.data, Receptor):
            return 1
        elif self.data is None:
            return 0
        else:
            return len(self.data)


class DbResult:
    """
    A class defining result objects for collector processes

    Attributes:
      id : result identifier
      data : list of original data records
      results: list of processed records
      data_pass: list of records that pass filtering for workers that split data before processing
      data_fail: list of records that failed filtering for workers that split data before processing
      valid : True if processing was successful and results should be written
      log : OrderedDict of log items
    """
    # Instantiation
    def __init__(self, key, records):
        self.id = key
        self.data = records
        self.results = None
        self.data_pass = records
        self.data_fail = None
        self.valid = False
        self.log = OrderedDict([('ID', key)])

    # Boolean evaluation
    def __bool__(self):
        return self.valid

    # Length evaluation
    def __len__(self):
        if isinstance(self.results, Receptor):
            return 1
        elif self.results is None:
            return 0
        else:
            return len(self.results)

    # Set data_count to number of data records
    @property
    def data_count(self):
        if isinstance(self.data, Receptor):
            return 1
        elif self.data is None:
            return 0
        else:
            return len(self.data)


def feedDbQueue(alive, data_queue, db_file, reader=AIRRReader, group_func=None, group_args={}):
    """
    Feeds the data queue with Ig records

    Arguments:
      alive : multiprocessing.Value boolean controlling whether processing continues
              if False exit process
      data_queue : multiprocessing.Queue to hold data for processing
      db_file : database file
      reader : database reader class
      group_func : function to use for grouping records
      group_args : dictionary of arguments to pass to group_func

    Returns:
      None
    """
    # Open input file and perform grouping
    try:
        # Iterate over records and assign groups
        db_handle = open(db_file, 'rt')
        db_iter = reader(db_handle)
        if group_func is not None:
            # import cProfile
            # prof = cProfile.Profile()
            # group_dict = prof.runcall(group_func, db_iter, **group_args)
            # prof.dump_stats('feed-%d.prof' % os.getpid())
            group_dict = group_func(db_iter, **group_args)
            group_iter = iter(group_dict.items())
        else:
            group_iter = ((r.sequence_id, r) for r in db_iter)
    except:
        alive.value = False
        raise

    # Add groups to data queue
    try:
        # Iterate over groups and feed data queue
        while alive.value:
            # Get data from queue
            if data_queue.full():  continue
            else:  data = next(group_iter, None)
            # Exit upon reaching end of iterator
            if data is None:  break

            # Feed queue
            data_queue.put(DbData(*data))
        else:
            sys.stderr.write('PID %s> Error in sibling process detected. Cleaning up.\n' \
                             % os.getpid())
            return None
    except:
        #sys.stderr.write('Exception in feeder queue feeding step\n')
        alive.value = False
        raise

    return None


def processDbQueue(alive, data_queue, result_queue, process_func, process_args={},
                   filter_func=None, filter_args={}):
    """
    Pulls from data queue, performs calculations, and feeds results queue

    Arguments:
      alive : multiprocessing.Value boolean controlling whether processing
            continues; when False function returns
      data_queue : multiprocessing.Queue holding data to process
      result_queue : multiprocessing.Queue to hold processed results
      process_func : function to use for processing sequences
      process_args : dictionary of arguments to pass to process_func
      filter_func : function to use for filtering sequences before processing
      filter_args : dictionary of arguments to pass to filter_func

    Returns:
      None
    """
    try:
        # Iterator over data queue until sentinel object reached
        while alive.value:
            # Get data from queue
            if data_queue.empty():  continue
            else:  data = data_queue.get()
            # Exit upon reaching sentinel
            if data is None:  break

            # Perform work
            if filter_func is None:
                result = process_func(data, **process_args)
            else:
                result = filter_func(data, **filter_args)
                result = process_func(result, **process_args)

            # Feed results to result queue
            result_queue.put(result)
        else:
            sys.stderr.write('PID %s> Error in sibling process detected. Cleaning up.\n' \
                             % os.getpid())
            return None
    except:
        alive.value = False
        printError('Processing data with ID: %s.' % str(data.id), exit=False)
        raise

    return None


def collectDbQueue(alive, result_queue, collect_queue, db_file, label, fields,
                   writer=AIRRWriter, out_file=None, out_args=default_out_args):
    """
    Pulls from results queue, assembles results and manages log and file IO

    Arguments:
      alive : multiprocessing.Value boolean controlling whether processing
              continues; when False function returns.
      result_queue : multiprocessing.Queue holding worker results.
      collect_queue : multiprocessing.Queue to store collector return values.
      db_file : database file name.
      label : task label used to tag the output files.
      fields : list of output fields.
      writer : writer class.
      out_file : output file name. Automatically generated from the input file if None.
      out_args : common output argument dictionary from parseCommonArgs.

    Returns:
      None: Adds a dictionary with key value pairs to collect_queue containing
           'log' defining a log object along with the 'pass' and 'fail' output file names.
    """
    # Wrapper for opening handles and writers
    def _open(x, f, writer=writer, label=label, out_file=out_file):
        if out_file is not None and x == 'pass':
            handle = open(out_file, 'w')
        else:
            handle = getOutputHandle(db_file,
                                     out_label='%s-%s' % (label, x),
                                     out_dir=out_args['out_dir'],
                                     out_name=out_args['out_name'],
                                     out_type=out_args['out_type'])
        return handle, writer(handle, fields=f)

    try:
        # Count input
        result_count = countDbFile(db_file)

        # Define log handle
        if out_args['log_file'] is None:
            log_handle = None
        else:
            log_handle = open(out_args['log_file'], 'w')
    except:
        alive.value = False
        raise

    try:
        # Initialize handles, writers and counters
        pass_handle, pass_writer = None, None
        fail_handle, fail_writer = None, None
        set_count, rec_count, pass_count, fail_count = 0, 0, 0, 0
        start_time = time()

        # Iterator over results queue until sentinel object reached
        while alive.value:
            # Get result from queue
            if result_queue.empty():  continue
            else:  result = result_queue.get()
            # Exit upon reaching sentinel
            if result is None:  break

            # Print progress for previous iteration
            printProgress(rec_count, result_count, 0.05, start_time=start_time)

            # Update counts for current iteration
            set_count += 1
            rec_count += result.data_count

            # Write log
            if result.log is not None:
                printLog(result.log, handle=log_handle)

            # Write output
            if result:
                # Write passing results
                pass_count += result.data_count
                try:
                    pass_writer.writeReceptor(result.results)
                except AttributeError:
                    # Open pass file and define writer object
                    pass_handle, pass_writer = _open('pass', fields)
                    pass_writer.writeReceptor(result.results)
            else:
                # Write failing data
                fail_count += result.data_count
                if out_args['failed']:
                    try:
                        fail_writer.writeReceptor(result.data)
                    except AttributeError:
                        # Open fail file and define writer object
                        fail_handle, fail_writer = _open('fail', fields)
                        fail_writer.writeReceptor(result.data)
        else:
            sys.stderr.write('PID %s> Error in sibling process detected. Cleaning up.\n' \
                             % os.getpid())
            return None

        # Print total counts
        printProgress(rec_count, result_count, 0.05, start_time=start_time)

        # Update log
        log = OrderedDict()
        log['OUTPUT'] = os.path.basename(pass_handle.name) if pass_handle is not None else None
        log['RECORDS'] = rec_count
        log['GROUPS'] = set_count
        log['PASS'] = pass_count
        log['FAIL'] = fail_count

        # Close file handles and generate return data
        collect_dict = {'log': log, 'pass': None, 'fail': None}
        if pass_handle is not None:
            collect_dict['pass'] = pass_handle.name
            pass_handle.close()
        if fail_handle is not None:
            collect_dict['fail'] = fail_handle.name
            fail_handle.close()
        if log_handle is not None:
            log_handle.close()
        collect_queue.put(collect_dict)
    except:
        alive.value = False
        raise

    return None


#!/usr/bin/env python3
"""
Create tab-delimited database file to store sequence alignment information
"""

# Info
__author__ = 'Namita Gupta, Jason Anthony Vander Heiden'
from changeo import __version__, __date__

# Imports
import os
import re
import csv
from argparse import ArgumentParser
from collections import OrderedDict
from textwrap import dedent
from time import time
from Bio import SeqIO

# Presto and changeo imports
from presto.Annotation import parseAnnotation
from presto.IO import countSeqFile, printLog, printMessage, printProgress, printError, printWarning, readSeqFile
from changeo.Defaults import default_format, default_out_args, default_imgt_id_len
from changeo.Commandline import CommonHelpFormatter, checkArgs, getCommonArgParser, parseCommonArgs
from changeo.Alignment import RegionDefinition, gapV
from changeo.Gene import buildGermline
from changeo.IO import countDbFile, extractIMGT, readGermlines, getFormatOperators, getOutputHandle, \
                       AIRRWriter, ChangeoWriter, IgBLASTReader, IgBLASTReaderAA, IMGTReader, IHMMuneReader, \
                       checkFields
from changeo.Receptor import ChangeoSchema, AIRRSchema

# 10X Receptor attributes
cellranger_base = ['cell', 'c_call', 'conscount', 'umicount']
cellranger_extended = ['cell', 'c_call', 'conscount', 'umicount',
                       'v_call_10x', 'd_call_10x', 'j_call_10x',
                       'junction_10x', 'junction_10x_aa']

def readCellRanger(cellranger_file, fields=cellranger_base):
    """
    Load a Cell Ranger annotation table

    Arguments:
      cellranger_file (str): path to the annotation file.
      fields (list): list of fields to keep.

    Returns:
      dict: dict of dicts with contig_id as the primary key.
    """
    # Mapping of 10X annotations to Receptor attributes
    cellranger_map = {'cell':  'barcode',
                      'c_call': 'c_gene',
                      'locus': 'chain',
                      'conscount': 'reads',
                      'umicount': 'umis',
                      'v_call_10x': 'v_gene',
                      'd_call_10x': 'd_gene',
                      'j_call_10x': 'j_gene',
                      'junction_10x': 'cdr3_nt',
                      'junction_10x_aa': 'cdr3'}

    # Function to parse individual fields
    def _parse(x):
        return '' if x == 'None' else x

    # Generate annotation dictionary
    ann_dict = {}
    with open(cellranger_file) as csv_file:
        # Detect delimiters
        dialect = csv.Sniffer().sniff(csv_file.readline())
        csv_file.seek(0)
        # Read in annotation file
        csv_reader = csv.DictReader(csv_file, dialect=dialect)

        # Generate annotation dictionary
        for row in csv_reader:
            ann_dict[row['contig_id']] = {f: _parse(row[cellranger_map[f]]) for f in fields}

    return ann_dict


def addGermline(receptor, references, amino_acid=False):
    """
    Add full length germline to Receptor object

    Arguments:
      receptor (changeo.Receptor.Receptor): Receptor object to modify.
      references (dict): dictionary of IMGT-gapped references sequences.
      amino_acid (bool): if True build amino acid germline, otherwise build nucleotide germline

    Returns:
      changeo.Receptor.Receptor: modified Receptor with the germline sequence added.
    """
    if amino_acid:
        __, germlines, __ = buildGermline(receptor, references, seq_field='sequence_aa_imgt',
                                          amino_acid=True)
        germline_seq = None if germlines is None else germlines['full']
        receptor.setField('germline_aa_imgt', germline_seq)
    else:
        __, germlines, __ = buildGermline(receptor, references, amino_acid=False)
        germline_seq = None if germlines is None else germlines['full']
        receptor.setField('germline_imgt', germline_seq)

    return receptor


def getIDforIMGT(seq_file, imgt_id_len=default_imgt_id_len):
    """
    Create a sequence ID translation using IMGT truncation.

    Arguments:
      seq_file (str): a fasta file of sequences input to IMGT.

    Returns:
      dict: a dictionary of with the IMGT truncated ID as the key and the full sequence description as the value.
    """

    # Create a sequence ID translation using IDs truncate up to space or 49 chars
    ids = {}
    for rec in readSeqFile(seq_file):
        if len(rec.description) <= imgt_id_len:
            id_key = rec.description
        else: # truncate and replace characters
            if imgt_id_len == 49: # 28 September 2021 (version 1.8.4)
                id_key = re.sub(r'\s|\t', '_', rec.description[:imgt_id_len])
            else: # older versions
                id_key = re.sub(r'\||\s|!|&|\*|<|>|\?', '_', rec.description[:imgt_id_len])
        ids.update({id_key: rec.description})

    return ids


def correctIMGTFields(receptor, references):
    """
    Add IMGT-numbering to IMGT fields in a Receptor object

    Arguments:
      receptor (changeo.Receptor.Receptor): Receptor object to modify.
      references (dict): dictionary of IMGT-gapped references sequences.

    Returns:
      changeo.Receptor.Receptor: modified Receptor with IMGT-gapped fields.
    """
    # Initialize update object
    imgt_dict = {'sequence_imgt': None,
                 'v_germ_start_imgt': None,
                 'v_germ_length_imgt': None,
                 'germline_imgt': None}

    # Check for necessary fields
    try:
        if not all([receptor.sequence_imgt,
                    receptor.v_germ_start_imgt,
                    receptor.v_germ_length_imgt,
                    receptor.v_call]):
            raise AttributeError
    except AttributeError:
        return None

    # Gap V region
    try:
        gapped = gapV(receptor.sequence_imgt,
                      receptor.v_germ_start_imgt,
                      receptor.v_germ_length_imgt,
                      receptor.v_call,
                      references)
    except KeyError as e:
        printWarning(e)
        return None

    # Verify IMGT-gapped sequence and junction concur
    # try:
    #     check = (receptor.junction == gapped['sequence_imgt'][309:(309 + receptor.junction_length)])
    # except TypeError:
    #     check = False
    # if not check:
    #     return None

    # Rebuild germline sequence
    receptor.setDict(gapped, parse=False)
    __, germlines, __ = buildGermline(receptor, references)
    # log, germlines, genes = buildGermline(receptor, references)
    # print(log)
    if germlines is not None:
        gapped['germline_imgt'] = germlines['full']
    else:
        return None

    # Update return object
    imgt_dict.update(gapped)

    return imgt_dict


def getSeqDict(seq_file):
    """
    Create a dictionary from a sequence file.

    Arguments:
      seq_file (str): sequence file.

    Returns:
      dict: sequence description as keys with Bio.SeqRecords as values.
    """
    seq_dict = SeqIO.to_dict(readSeqFile(seq_file), key_function=lambda x: x.description)

    return seq_dict


def writeDb(records, fields, aligner_file, total_count, id_dict=None, annotations=None,
            amino_acid=False, validate='strict', asis_id=True, regions='default',
            writer=AIRRWriter, out_file=None, out_args=default_out_args):
    """
    Writes parsed records to an output file

    Arguments:
      records (iter): a iterator of Receptor objects containing alignment data.
      fields (list): a list of ordered field names to write.
      aligner_file (str): input file name.
      total_count (int): number of records (for progress bar).
      id_dict (dict): a dictionary of the truncated sequence ID mapped to the full sequence ID.
      annotations (dict): additional annotation dictionary.
      amino_acid (bool): if True do verification on amino acid fields.
      validate (str): validation criteria for passing records; one of 'strict', 'gentle', or 'partial'.
      asis_id (bool): if ID is to be parsed for pRESTO output with default delimiters.
      regions (str): name of the IMGT FWR/CDR region definitions to use.
      writer (changeo.IO.TSVWriter): writer class.
      out_file (str): output file name. Automatically generated from the input file if None.
      out_args (dict): common output argument dictionary from parseCommonArgs.

    Returns:
      None
    """
    # Wrapper for opening handles and writers
    def _open(x, f, writer=writer, out_file=out_file):
        if out_file is not None and x == 'pass':
            handle = open(out_file, 'w')
        else:
            handle = getOutputHandle(aligner_file,
                                     out_label='db-%s' % x,
                                     out_dir=out_args['out_dir'],
                                     out_name=out_args['out_name'],
                                     out_type=out_args['out_type'])
        return handle, writer(handle, fields=f)

    # Function to convert fasta header annotations to changeo columns
    def _changeo(f, header):
        h = [ChangeoSchema.fromReceptor(x) for x in header if x.upper() not in f]
        f.extend(h)
        return f

    def _airr(f, header):
        h = [AIRRSchema.fromReceptor(x) for x in header if x.lower() not in f]
        f.extend(h)
        return f

    # Function to verify IMGT-gapped sequence and junction concur
    def _imgt_check(rec):
        try:
            if amino_acid:
                rd = RegionDefinition(rec.junction_aa_length, amino_acid=amino_acid, definition=regions)
                x, y = rd.positions['junction']
                check = (rec.junction_aa == rec.sequence_aa_imgt[x:y])
            else:
                rd = RegionDefinition(rec.junction_length, amino_acid=amino_acid, definition=regions)
                x, y = rd.positions['junction']
                check = (rec.junction == rec.sequence_imgt[x:y])
        except (TypeError, AttributeError):
            check = False
        return check

    # Default function to check for valid records
    def _gentle(rec):
        if amino_acid:
            valid = [rec.v_call and rec.v_call != 'None',
                     rec.j_call and rec.j_call != 'None',
                     rec.functional is not None,
                     rec.sequence_aa_imgt,
                     rec.junction_aa]
        else:
            valid = [rec.v_call and rec.v_call != 'None',
                     rec.j_call and rec.j_call != 'None',
                     rec.functional is not None,
                     rec.sequence_imgt,
                     rec.junction]
        return all(valid)

    # Function to check for valid records strictly
    def _strict(rec):
        if amino_acid:
            valid = [rec.v_call and rec.v_call != 'None',
                     rec.j_call and rec.j_call != 'None',
                     rec.functional is not None,
                     rec.sequence_aa_imgt,
                     rec.junction_aa,
                     _imgt_check(rec)]
        else:
            valid = [rec.v_call and rec.v_call != 'None',
                     rec.j_call and rec.j_call != 'None',
                     rec.functional is not None,
                     rec.sequence_imgt,
                     rec.junction,
                     _imgt_check(rec)]
        return all(valid)

    # Function to check for valid records loosely
    def _partial(rec):
        valid = [rec.v_call and rec.v_call != 'None',
                 rec.d_call and rec.d_call != 'None',
                 rec.j_call and rec.j_call != 'None']
        return any(valid)

    # Set writer class and annotation conversion function
    if writer == ChangeoWriter:
        _annotate = _changeo
    elif writer == AIRRWriter:
        _annotate = _airr
    else:
        printError('Invalid output writer.')

    # Set pass criteria
    validate_map = {'strict': _strict, 'gentle': _gentle, 'partial': _partial}
    _pass = validate_map.get(validate, _gentle)

    # Define log handle
    if out_args['log_file'] is None:
        log_handle = None
    else:
        log_handle = open(out_args['log_file'], 'w')

    # Initialize handles, writers and counters
    pass_handle, pass_writer = None, None
    fail_handle, fail_writer = None, None
    pass_count, fail_count = 0, 0
    start_time = time()

    # Validate and write output
    printProgress(0, total_count, 0.05, start_time=start_time)
    for i, record in enumerate(records, start=1):
        # Replace sequence description with full string, if required
        if id_dict is not None and record.sequence_id in id_dict:
            record.sequence_id = id_dict[record.sequence_id]

        # Parse sequence description into new columns
        if not asis_id:
            try:
                ann_raw = parseAnnotation(record.sequence_id)
                record.sequence_id = ann_raw.pop('ID')

                # Convert to Receptor fields
                ann_parsed = OrderedDict()
                for k, v in ann_raw.items():
                    ann_parsed[ChangeoSchema.toReceptor(k)] = v

                # Add annotations to Receptor and update field list
                record.setDict(ann_parsed, parse=True)
                if i == 1:  fields = _annotate(fields, ann_parsed.keys())
            except IndexError:
                # Could not parse pRESTO-style annotations so fall back to no parse
                asis_id = True
                printWarning('Sequence annotation format not recognized. Sequence headers will not be parsed.')

        # Add supplemental annotation fields
        # if _append_table is not None:
        #     record.setDict(_append_table(record.sequence_id), parse=True)
        if annotations is not None:
            record.setDict(annotations[record.sequence_id], parse=True)
            if i == 1:  fields = _annotate(fields, annotations[record.sequence_id].keys())

        # Count pass or fail and write to appropriate file
        if _pass(record):
            pass_count += 1
            # Write row to pass file
            try:
                pass_writer.writeReceptor(record)
            except AttributeError:
                # Open pass file and writer
                pass_handle, pass_writer = _open('pass', fields)
                pass_writer.writeReceptor(record)
        else:
            fail_count += 1
            # Write row to fail file if specified
            if out_args['failed']:
                try:
                    fail_writer.writeReceptor(record)
                except AttributeError:
                    # Open fail file and writer
                    fail_handle, fail_writer = _open('fail', fields)
                    fail_writer.writeReceptor(record)

        # Write log
        if log_handle is not None:
            log = OrderedDict([('ID', record.sequence_id),
                               ('V_CALL', record.v_call),
                               ('D_CALL', record.d_call),
                               ('J_CALL', record.j_call),
                               ('PRODUCTIVE', record.functional)])
            if not _imgt_check(record) and not amino_acid:
                log['ERROR'] = 'Junction does not match the sequence starting at position 310 in the IMGT numbered V(D)J sequence.'
            printLog(log, log_handle)

        # Print progress
        printProgress(i, total_count, 0.05, start_time=start_time)

    # Print console log
    log = OrderedDict()
    log['OUTPUT'] = os.path.basename(pass_handle.name) if pass_handle is not None else None
    log['PASS'] = pass_count
    log['FAIL'] = fail_count
    log['END'] = 'MakeDb'
    printLog(log)

    # Close file handles
    output = {'pass': None, 'fail': None}
    if pass_handle is not None:
        output['pass'] = pass_handle.name
        pass_handle.close()
    if fail_handle is not None:
        output['fail'] = fail_handle.name
        fail_handle.close()

    return output


def parseIMGT(aligner_file, seq_file=None, repo=None, cellranger_file=None, validate='strict', asis_id=True,
              extended=False, format=default_format, out_file=None, out_args=default_out_args,
              imgt_id_len=default_imgt_id_len):
    """
    Main for IMGT aligned sample sequences.

    Arguments:
      aligner_file (str): zipped file or unzipped folder output by IMGT.
      seq_file (str): FASTA file input to IMGT (from which to get seqID).
      repo (str): folder with germline repertoire files.
      validate (str): validation criteria for passing records; one of 'strict', 'gentle', or 'partial'.
      asis_id (bool): if ID is to be parsed for pRESTO output with default delimiters.
      extended (bool): if True add alignment score, FWR, CDR and junction fields to output file.
      format (str): output format. one of 'changeo' or 'airr'.
      out_file (str): output file name. Automatically generated from the input file if None.
      out_args (dict): common output argument dictionary from parseCommonArgs.
      imgt_id_len (int): maximum character length of sequence identifiers reported by IMGT/HighV-QUEST.

    Returns:
      dict: names of the 'pass' and 'fail' output files.
    """
    # Print parameter info
    log = OrderedDict()
    log['START'] = 'MakeDb'
    log['COMMAND'] = 'imgt'
    log['ALIGNER_FILE'] = aligner_file
    log['SEQ_FILE'] = os.path.basename(seq_file) if seq_file else ''
    log['ASIS_ID'] = asis_id
    log['VALIDATE'] = validate
    log['EXTENDED'] = extended
    printLog(log)

    start_time = time()
    printMessage('Loading files', start_time=start_time, width=20)

    # Extract IMGT files
    temp_dir, imgt_files = extractIMGT(aligner_file)

    # Count records in IMGT files
    total_count = countDbFile(imgt_files['summary'])

    # Get (parsed) IDs from fasta file submitted to IMGT
    id_dict = getIDforIMGT(seq_file, imgt_id_len) if seq_file else {}

    # Load supplementary annotation table
    if cellranger_file is not None:
        f = cellranger_extended if extended else cellranger_base
        annotations = readCellRanger(cellranger_file, fields=f)
    else:
        annotations = None

    printMessage('Done', start_time=start_time, end=True, width=20)

    # Define format operators
    try:
        __, writer, schema = getFormatOperators(format)
    except ValueError:
        printError('Invalid format %s.' % format)
    out_args['out_type'] = schema.out_type

    # Define output fields
    fields = list(schema.required)
    if extended:
        custom = IMGTReader.customFields(scores=True, regions=True, junction=True, schema=schema)
        fields.extend(custom)

    # Parse IMGT output and write db
    with open(imgt_files['summary'], 'r') as summary_handle, \
            open(imgt_files['gapped'], 'r') as gapped_handle, \
            open(imgt_files['ntseq'], 'r') as ntseq_handle, \
            open(imgt_files['junction'], 'r') as junction_handle:

        # Open parser
        parse_iter = IMGTReader(summary_handle, gapped_handle, ntseq_handle, junction_handle)

        # Add germline sequence
        if repo is None:
            germ_iter = parse_iter
        else:
            references = readGermlines(repo)
            # Check for IMGT-gaps in germlines
            if all('...' not in x for x in references.values()):
                printWarning('Germline reference sequences do not appear to contain IMGT-numbering spacers. Results may be incorrect.')
            germ_iter = (addGermline(x, references) for x in parse_iter)
        # Write db
        output = writeDb(germ_iter, fields=fields, aligner_file=aligner_file, total_count=total_count,
                         annotations=annotations, id_dict=id_dict, asis_id=asis_id, validate=validate,
                         writer=writer, out_file=out_file, out_args=out_args)

    # Cleanup temp directory
    temp_dir.cleanup()

    return output


def parseIgBLAST(aligner_file, seq_file, repo, amino_acid=False, cellranger_file=None, validate='strict',
                 asis_id=True, asis_calls=False, extended=False, regions='default', infer_junction=False,
                 format='changeo', out_file=None, out_args=default_out_args):
    """
    Main for IgBLAST aligned sample sequences.

    Arguments:
      aligner_file (str): IgBLAST output file to process.
      seq_file (str): fasta file input to IgBlast (from which to get sequence).
      repo (str): folder with germline repertoire files.
      amino_acid (bool): if True then the IgBLAST output files are results from igblastp. igblastn is assumed if False.
      validate (str): validation criteria for passing records; one of 'strict', 'gentle', or 'partial'.
      asis_id (bool): if ID is to be parsed for pRESTO output with default delimiters.
      asis_calls (bool): if True do not parse gene calls for allele names.
      extended (bool): if True add alignment scores, FWR regions, and CDR regions to the output.
      regions (str): name of the IMGT FWR/CDR region definitions to use.
      infer_junction (bool): if True, infer the junction sequence, if not reported by IgBLAST.
      format (str): output format. one of 'changeo' or 'airr'.
      out_file (str): output file name. Automatically generated from the input file if None.
      out_args (dict): common output argument dictionary from parseCommonArgs.

    Returns:
      dict: names of the 'pass' and 'fail' output files.
    """
    # Print parameter info
    log = OrderedDict()
    log['START'] = 'MakeDB'
    log['COMMAND'] = 'igblast-aa' if amino_acid else 'igblast'
    log['ALIGNER_FILE'] = os.path.basename(aligner_file)
    log['SEQ_FILE'] = os.path.basename(seq_file)
    log['ASIS_ID'] = asis_id
    log['ASIS_CALLS'] = asis_calls
    log['VALIDATE'] = validate
    log['EXTENDED'] = extended
    log['INFER_JUNCTION'] = infer_junction
    printLog(log)

    # Set amino acid conditions
    if amino_acid:
        format = '%s-aa' % format
        parser = IgBLASTReaderAA
    else:
        parser = IgBLASTReader

    # Start
    start_time = time()
    printMessage('Loading files', start_time=start_time, width=20)

    # Count records in sequence file
    total_count = countSeqFile(seq_file)

    # Get input sequence dictionary
    seq_dict = getSeqDict(seq_file)

    # Create germline repo dictionary
    references = readGermlines(repo, asis=asis_calls)

    # Load supplementary annotation table
    if cellranger_file is not None:
        f = cellranger_extended if extended else cellranger_base
        annotations = readCellRanger(cellranger_file, fields=f)
    else:
        annotations = None

    printMessage('Done', start_time=start_time, end=True, width=20)

    # Check for IMGT-gaps in germlines
    if all('...' not in x for x in references.values()):
        printWarning('Germline reference sequences do not appear to contain IMGT-numbering spacers. Results may be incorrect.')

    # Define format operators
    try:
        __, writer, schema = getFormatOperators(format)
    except ValueError:
        printError('Invalid format %s.' % format)
    out_args['out_type'] = schema.out_type

    # Define output fields
    fields = list(schema.required)
    if extended:
        custom = parser.customFields(schema=schema)
        fields.extend(custom)

    # Parse and write output
    with open(aligner_file, 'r') as f:
        parse_iter = parser(f, seq_dict, references, regions=regions, asis_calls=asis_calls, infer_junction=infer_junction)
        germ_iter = (addGermline(x, references, amino_acid=amino_acid) for x in parse_iter)
        output = writeDb(germ_iter, fields=fields, aligner_file=aligner_file, total_count=total_count,
                         annotations=annotations, amino_acid=amino_acid, validate=validate, asis_id=asis_id,
                         regions=regions, writer=writer, out_file=out_file, out_args=out_args)

    return output


def parseIHMM(aligner_file, seq_file, repo, cellranger_file=None, validate='strict', asis_id=True,
              extended=False, format=default_format, out_file=None, out_args=default_out_args):
    """
    Main for iHMMuneAlign aligned sample sequences.

    Arguments:
      aligner_file (str): iHMMune-Align output file to process.
      seq_file (str): fasta file input to iHMMuneAlign (from which to get sequence).
      repo (str): folder with germline repertoire files.
      validate (str): validation criteria for passing records; one of 'strict', 'gentle', or 'partial'.
      asis_id (bool): if ID is to be parsed for pRESTO output with default delimiters.
      extended (bool): if True parse alignment scores, FWR and CDR region fields.
      format (str): output format. One of 'changeo' or 'airr'.
      out_file (str): output file name. Automatically generated from the input file if None.
      out_args (dict): common output argument dictionary from parseCommonArgs.

    Returns:
      dict: names of the 'pass' and 'fail' output files.
    """
    # Print parameter info
    log = OrderedDict()
    log['START'] = 'MakeDB'
    log['COMMAND'] = 'ihmm'
    log['ALIGNER_FILE'] = os.path.basename(aligner_file)
    log['SEQ_FILE'] = os.path.basename(seq_file)
    log['ASIS_ID'] = asis_id
    log['VALIDATE'] = validate
    log['EXTENDED'] = extended
    printLog(log)

    start_time = time()
    printMessage('Loading files', start_time=start_time, width=20)

    # Count records in sequence file
    total_count = countSeqFile(seq_file)

    # Get input sequence dictionary
    seq_dict = getSeqDict(seq_file)

    # Create germline repo dictionary
    references = readGermlines(repo)

    # Load supplementary annotation table
    if cellranger_file is not None:
        f = cellranger_extended if extended else cellranger_base
        annotations = readCellRanger(cellranger_file, fields=f)
    else:
        annotations = None

    printMessage('Done', start_time=start_time, end=True, width=20)

    # Check for IMGT-gaps in germlines
    if all('...' not in x for x in references.values()):
        printWarning('Germline reference sequences do not appear to contain IMGT-numbering spacers. Results may be incorrect.')

    # Define format operators
    try:
        __, writer, schema = getFormatOperators(format)
    except ValueError:
        printError('Invalid format %s.' % format)
    out_args['out_type'] = schema.out_type

    # Define output fields
    fields = list(schema.required)
    if extended:
        custom = IHMMuneReader.customFields(scores=True, regions=True, schema=schema)
        fields.extend(custom)

    # Parse and write output
    with open(aligner_file, 'r') as f:
        parse_iter = IHMMuneReader(f, seq_dict, references)
        germ_iter = (addGermline(x, references) for x in parse_iter)
        output = writeDb(germ_iter, fields=fields, aligner_file=aligner_file, total_count=total_count,
                        annotations=annotations, asis_id=asis_id, validate=validate,
                        writer=writer, out_file=out_file, out_args=out_args)

    return output



def numberAIRR(aligner_file, repo=None, format=default_format,
               out_file=None, out_args=default_out_args):
    """
    Inserts IMGT numbering into V fields

    Arguments:
      aligner_file (str): AIRR Rearrangement file from the alignment tool.
      repo (str): folder with germline repertoire files. If None, do not updated alignment columns with IMGT gaps.
      format (str): output format.
      out_file (str): output file name. Automatically generated from the input file if None.
      out_args (dict): common output argument dictionary from parseCommonArgs.

    Returns:
      str: output file name.
    """
    log = OrderedDict()
    log['START'] = 'MakeDb'
    log['COMMAND'] = 'number'
    log['ALIGNER_FILE'] = os.path.basename(aligner_file)
    printLog(log)

    # Define format operators
    try:
        reader, writer, schema = getFormatOperators(format)
    except ValueError:
        printError('Invalid format %s.' % format)

    # Open input
    db_handle = open(aligner_file, 'rt')
    db_iter = reader(db_handle)

    # Define log handle
    if out_args['log_file'] is None:
        log_handle = None
    else:
        log_handle = open(out_args['log_file'], 'w')

    # Check for required columns
    try:
        required = ['sequence_imgt', 'v_germ_start_imgt']
        checkFields(required, db_iter.fields, schema=schema)
    except LookupError as e:
        printError(e)

    # Load references
    reference_dict = readGermlines(repo)

    # Check for IMGT-gaps in germlines
    if all('...' not in x for x in reference_dict.values()):
        printWarning('Germline reference sequences do not appear to contain IMGT-numbering spacers. Results may be incorrect.')

    # Open output writer
    if out_file is not None:
        pass_handle = open(out_file, 'w')
    else:
        pass_handle = getOutputHandle(aligner_file, out_label='db-pass', out_dir=out_args['out_dir'],
                                        out_name=out_args['out_name'], out_type=schema.out_type)
    pass_writer = writer(pass_handle, fields=db_iter.fields)

    if out_args['failed']:
        fail_handle = getOutputHandle(aligner_file, out_label='db-fail', out_dir=out_args['out_dir'],
                                        out_name=out_args['out_name'], out_type=schema.out_type)
        fail_writer = writer(fail_handle, fields=db_iter.fields)

    # Count records
    result_count = countDbFile(aligner_file)

    # Iterate over records
    start_time = time()
    rec_count = pass_count = fail_count= 0
    for rec in db_iter:
        # Print progress for previous iteration
        printProgress(rec_count, result_count, 0.05, start_time=start_time)
        rec_count += 1
        # Update IMGT fields
        imgt_dict = correctIMGTFields(rec, reference_dict)
        # Write records
        if imgt_dict is not None:
            pass_count += 1
            rec.setDict(imgt_dict, parse=False)
            pass_writer.writeReceptor(rec)
        else:
            fail_count += 1
            # Write row to fail file if specified
            if out_args['failed']:
                fail_writer.writeReceptor(rec)
        # Write log
        if log_handle is not None:
            log = OrderedDict([('ID', rec.sequence_id),
                                ('V_CALL', rec.v_call),
                                ('D_CALL', rec.d_call),
                                ('J_CALL', rec.j_call),
                                ('PRODUCTIVE', rec.functional)])
            printLog(log, log_handle)

    # Print counts
    printProgress(rec_count, result_count, 0.05, start_time=start_time)
    log = OrderedDict()
    log['OUTPUT'] = os.path.basename(pass_handle.name)
    log['RECORDS'] = rec_count
    log['PASS'] = pass_count
    log['FAIL'] = rec_count - pass_count
    log['END'] = 'MakeDb'
    printLog(log)

    # Close file handles
    pass_handle.close()
    db_handle.close()

    return pass_handle.name


def getArgParser():
    """
    Defines the ArgumentParser.

    Returns:
      argparse.ArgumentParser
    """
    fields = dedent(
             '''
              output files:
                  db-pass
                      database of alignment records with functionality information,
                      V and J calls, and a junction region.
                  db-fail
                      database with records that fail due to no productivity information,
                      no gene V assignment, no J assignment, or no junction region.

              universal output fields:
                 sequence_id, sequence, sequence_alignment, germline_alignment,
                 rev_comp, productive, stop_codon, vj_in_frame, locus,
                 v_call, d_call, j_call, c_call, junction, junction_length, junction_aa,
                 v_sequence_start, v_sequence_end, v_germline_start, v_germline_end,
                 d_sequence_start, d_sequence_end, d_germline_start, d_germline_end,
                 j_sequence_start, j_sequence_end, j_germline_start, j_germline_end,
                 np1_length, np2_length, fwr1, fwr2, fwr3, fwr4, cdr1, cdr2, cdr3

              imgt specific output fields:
                  n1_length, n2_length, p3v_length, p5d_length, p3d_length, p5j_length,
                  d_frame, v_score, v_identity, d_score, d_identity, j_score, j_identity

              igblast specific output fields:
                  v_score, v_identity, v_support, v_cigar,
                  d_score, d_identity, d_support, d_cigar,
                  j_score, j_identity, j_support, j_cigar

              ihmm specific output fields:
                  vdj_score

              10x specific output fields:
                  cell_id, consensus_count, umi_count,
                  v_call_10x, d_call_10x, j_call_10x,
                  junction_10x, junction_10x_aa
              ''')

    # Define ArgumentParser
    parser = ArgumentParser(description=__doc__, epilog=fields,
                            formatter_class=CommonHelpFormatter, add_help=False)
    group_help = parser.add_argument_group('help')
    group_help.add_argument('--version', action='version',
                            version='%(prog)s:' + ' %s %s' %(__version__, __date__))
    group_help.add_argument('-h', '--help', action='help', help='show this help message and exit')
    subparsers = parser.add_subparsers(title='subcommands', dest='command',
                                       help='Aligner used', metavar='')
    # TODO:  This is a temporary fix for Python issue 9253
    subparsers.required = True

    # Parent parser
    parser_parent = getCommonArgParser(db_in=False)

    # igblastn output parser
    parser_igblast = subparsers.add_parser('igblast', parents=[parser_parent],
                                           formatter_class=CommonHelpFormatter, add_help=False,
                                           help='Process igblastn output.',
                                           description='Process igblastn output.')
    group_igblast = parser_igblast.add_argument_group('aligner parsing arguments')
    group_igblast.add_argument('-i', nargs='+', action='store', dest='aligner_files', required=True,
                                help='''IgBLAST output files in format 7 with query sequence
                                     (igblastn argument \'-outfmt "7 std qseq sseq btop"\').''')
    group_igblast.add_argument('-r', nargs='+', action='store', dest='repo', required=True,
                                help='''List of folders and/or fasta files containing
                                     the same germline set used in the IgBLAST alignment. These
                                     reference sequences must contain IMGT-numbering spacers (gaps)
                                     in the V segment.''')
    group_igblast.add_argument('-s', action='store', nargs='+', dest='seq_files',
                                required=True,
                                help='''List of input FASTA files (with .fasta, .fna or .fa
                                     extension), containing sequences.''')
    group_igblast.add_argument('--10x', action='store', nargs='+', dest='cellranger_files',
                                help='''Table file containing 10X annotations (with .csv or .tsv
                                     extension).''')
    group_igblast.add_argument('--asis-id', action='store_true', dest='asis_id',
                                help='''Specify to prevent input sequence headers from being parsed
                                     to add new columns to database. Parsing of sequence headers requires
                                     headers to be in the pRESTO annotation format, so this should be specified
                                     when sequence headers are incompatible with the pRESTO annotation scheme.
                                     Note, unrecognized header formats will default to this behavior.''')
    group_igblast.add_argument('--asis-calls', action='store_true', dest='asis_calls',
                                help='''Specify to prevent gene calls from being parsed into standard allele names
                                     in both the IgBLAST output and reference database. Note, this requires
                                     the sequence identifiers in the reference sequence set and the IgBLAST
                                     database to be exact string matches.''')
    group_igblast.add_argument('--extended', action='store_true', dest='extended',
                               help='''Specify to include additional aligner specific fields in the output.
                                    Adds <vdj>_score, <vdj>_identity, <vdj>_support, <vdj>_cigar,
                                    fwr1, fwr2, fwr3, fwr4, cdr1, cdr2 and cdr3.''')
    group_igblast.add_argument('--regions', action='store', dest='regions',
                               choices=('default', 'rhesus-igl'), default='default',
                               help='''IMGT CDR and FWR boundary definition to use.''')
    group_igblast.add_argument('--infer-junction', action='store_true', dest='infer_junction',
                               help='''Infer the junction sequence. For use with IgBLAST v1.6.0 or older,
                                    prior to the addition of IMGT-CDR3 inference.''')
    group_igblast_validate = group_igblast.add_mutually_exclusive_group(required=False)
    # group_igblast_validate.add_argument('--strict', action='store_const', const='strict', dest='validate',
    #                                     help='''By default, passing records must contain valid values for the
    #                                          V gene, J gene, junction region, and productivity call. If specified,
    #                                          this argument adds the additional requirement that the junction region must
    #                                          start at position 310 in the IMGT-numbered sequence.''')
    group_igblast_validate.add_argument('--partial', action='store_const', const='partial', dest='validate',
                                        help='''If specified, include incomplete V(D)J alignments in
                                             the pass file instead of the fail file. An incomplete alignment
                                             is defined as a record that is missing a V gene assignment,
                                             J gene assignment, junction region, or productivity call.''')
    parser_igblast.set_defaults(func=parseIgBLAST, amino_acid=False, validate='strict')

    # igblastp output parser
    parser_igblast_aa = subparsers.add_parser('igblast-aa', parents=[parser_parent],
                                           formatter_class=CommonHelpFormatter, add_help=False,
                                           help='Process igblastp output.',
                                           description='Process igblastp output.')
    group_igblast_aa = parser_igblast_aa.add_argument_group('aligner parsing arguments')
    group_igblast_aa.add_argument('-i', nargs='+', action='store', dest='aligner_files', required=True,
                                  help='''IgBLAST output files in format 7 with query sequence
                                       (igblastp argument \'-outfmt "7 std qseq sseq btop"\').''')
    group_igblast_aa.add_argument('-r', nargs='+', action='store', dest='repo', required=True,
                                  help='''List of folders and/or fasta files containing
                                       the same germline set used in the IgBLAST alignment. These
                                       reference sequences must contain IMGT-numbering spacers (gaps)
                                       in the V segment.''')
    group_igblast_aa.add_argument('-s', action='store', nargs='+', dest='seq_files', required=True,
                                  help='''List of input FASTA files (with .fasta, .fna or .fa
                                       extension), containing sequences.''')
    group_igblast_aa.add_argument('--10x', action='store', nargs='+', dest='cellranger_files',
                                  help='''Table file containing 10X annotations (with .csv or .tsv extension).''')
    group_igblast_aa.add_argument('--asis-id', action='store_true', dest='asis_id',
                                  help='''Specify to prevent input sequence headers from being parsed
                                       to add new columns to database. Parsing of sequence headers requires
                                       headers to be in the pRESTO annotation format, so this should be specified
                                       when sequence headers are incompatible with the pRESTO annotation scheme.
                                       Note, unrecognized header formats will default to this behavior.''')
    group_igblast_aa.add_argument('--asis-calls', action='store_true', dest='asis_calls',
                                  help='''Specify to prevent gene calls from being parsed into standard allele names
                                       in both the IgBLAST output and reference database. Note, this requires
                                       the sequence identifiers in the reference sequence set and the IgBLAST
                                       database to be exact string matches.''')
    group_igblast_aa.add_argument('--extended', action='store_true', dest='extended',
                                  help='''Specify to include additional aligner specific fields in the output.
                                       Adds v_score, v_identity, v_support, v_cigar, fwr1, fwr2, fwr3, cdr1 and cdr2.''')
    group_igblast_aa.add_argument('--regions', action='store', dest='regions',
                                  choices=('default', 'rhesus-igl'), default='default',
                                  help='''IMGT CDR and FWR boundary definition to use.''')
    parser_igblast_aa.set_defaults(func=parseIgBLAST, amino_acid=True, validate='strict')


    # IMGT aligner
    parser_imgt = subparsers.add_parser('imgt', parents=[parser_parent],
                                        formatter_class=CommonHelpFormatter, add_help=False,
                                        help='''Process IMGT/HighV-Quest output
                                             (does not work with V-QUEST).''',
                                        description='''Process IMGT/HighV-Quest output
                                             (does not work with V-QUEST).''')
    group_imgt = parser_imgt.add_argument_group('aligner parsing arguments')
    group_imgt.add_argument('-i', nargs='+', action='store', dest='aligner_files', required=True,
                            help='''Either zipped IMGT output files (.zip or .txz) or a
                                 folder containing unzipped IMGT output files (which must
                                 include 1_Summary, 2_IMGT-gapped, 3_Nt-sequences,
                                 and 6_Junction).''')
    group_imgt.add_argument('-s', nargs='*', action='store', dest='seq_files', required=False,
                            help='''List of FASTA files (with .fasta, .fna or .fa
                                  extension) that were submitted to IMGT/HighV-QUEST.
                                  If unspecified, sequence identifiers truncated by IMGT/HighV-QUEST
                                  will not be corrected.''')
    group_imgt.add_argument('-r', nargs='+', action='store', dest='repo', required=False,
                            help='''List of folders and/or fasta files containing
                                 the germline sequence set used by IMGT/HighV-QUEST.
                                 These reference sequences must contain IMGT-numbering spacers (gaps)
                                 in the V segment. If unspecified, the germline sequence reconstruction
                                 will not be included in the output.''')
    group_imgt.add_argument('--10x', action='store', nargs='+', dest='cellranger_files',
                            help='''Table file containing 10X annotations (with .csv or .tsv
                                 extension).''')
    group_imgt.add_argument('--extended', action='store_true', dest='extended',
                            help='''Specify to include additional aligner specific fields in the output.
                                 Adds <vdj>_score, <vdj>_identity>, fwr1, fwr2, fwr3, fwr4,
                                 cdr1, cdr2, cdr3, n1_length, n2_length, p3v_length, p5d_length,
                                 p3d_length, p5j_length and d_frame.''')
    group_imgt.add_argument('--asis-id', action='store_true', dest='asis_id',
                            help='''Specify to prevent input sequence headers from being parsed
                                 to add new columns to database. Parsing of sequence headers requires
                                 headers to be in the pRESTO annotation format, so this should be specified
                                 when sequence headers are incompatible with the pRESTO annotation scheme.
                                 Note, unrecognized header formats will default to this behavior.''')
    group_imgt.add_argument('--imgt-id-len', action='store', dest='imgt_id_len', type=int,
                            default=default_imgt_id_len,
                            help='''The maximum character length of sequence identifiers reported by IMGT/HighV-QUEST.
                                 Specify 50 if the IMGT files (-i) were generated with an IMGT/HighV-QUEST version older
                                 than 1.8.3 (May 7, 2021).''')
    group_imgt_validate = group_imgt.add_mutually_exclusive_group(required=False)
    # group_imgt_validate.add_argument('--strict', action='store_const', const='strict', dest='validate',
    #                                  help='''By default, passing records must contain valid values for the
    #                                       V gene, J gene, junction region, and productivity call. If specified,
    #                                       this argument adds the additional requirement that the junction region must
    #                                       start at position 310 in the IMGT-numbered sequence.''')
    group_imgt_validate.add_argument('--partial', action='store_const', const='partial', dest='validate',
                                     help='''If specified, include incomplete V(D)J alignments in
                                          the pass file instead of the fail file. An incomplete alignment
                                          is defined as a record that is missing a V gene assignment,
                                          J gene assignment, junction region, or productivity call.''')
    parser_imgt.set_defaults(func=parseIMGT, validate='strict')

    # iHMMuneAlign Aligner
    parser_ihmm = subparsers.add_parser('ihmm', parents=[parser_parent],
                                        formatter_class=CommonHelpFormatter, add_help=False,
                                        help='Process iHMMune-Align output.',
                                        description='Process iHMMune-Align output.')
    group_ihmm = parser_ihmm.add_argument_group('aligner parsing arguments')
    group_ihmm.add_argument('-i', nargs='+', action='store', dest='aligner_files',
                             required=True,
                             help='''iHMMune-Align output file.''')
    group_ihmm.add_argument('-r', nargs='+', action='store', dest='repo', required=True,
                             help='''List of folders and/or FASTA files containing
                                   the set of germline sequences used by iHMMune-Align. These
                                   reference sequences must contain IMGT-numbering spacers (gaps)
                                   in the V segment.''')
    group_ihmm.add_argument('-s', action='store', nargs='+', dest='seq_files',
                             required=True,
                             help='''List of input FASTA files (with .fasta, .fna or .fa
                                  extension) containing sequences.''')
    group_ihmm.add_argument('--10x', action='store', nargs='+', dest='cellranger_files',
                                help='''Table file containing 10X annotations (with .csv or .tsv
                                     extension).''')
    group_ihmm.add_argument('--asis-id', action='store_true', dest='asis_id',
                             help='''Specify to prevent input sequence headers from being parsed
                                  to add new columns to database. Parsing of sequence headers requires
                                  headers to be in the pRESTO annotation format, so this should be specified
                                  when sequence headers are incompatible with the pRESTO annotation scheme.
                                  Note, unrecognized header formats will default to this behavior.''')
    group_ihmm.add_argument('--extended', action='store_true', dest='extended',
                             help='''Specify to include additional aligner specific fields in the output.
                                  Adds the path score of the iHMMune-Align hidden Markov model as vdj_score;
                                  adds fwr1, fwr2, fwr3, fwr4, cdr1, cdr2 and cdr3.''')
    group_ihmm_validate = group_ihmm.add_mutually_exclusive_group(required=False)
    # group_ihmm_validate.add_argument('--strict', action='store_const', const='strict', dest='validate',
    #                                  help='''By default, passing records must contain valid values for the
    #                                       V gene, J gene, junction region, and productivity call. If specified,
    #                                       this argument adds the additional requirement that the junction region must
    #                                       start at position 310 in the IMGT-numbered sequence.''')
    group_ihmm_validate.add_argument('--partial', action='store_const', const='partial', dest='validate',
                                     help='''If specified, include incomplete V(D)J alignments in
                                          the pass file instead of the fail file. An incomplete alignment
                                          is defined as a record that is missing a V gene assignment,
                                          J gene assignment, junction region, or productivity call.''')
    parser_ihmm.set_defaults(func=parseIHMM, validate='strict')

    # Subparser to normalize AIRR file with IMGT-numbering
    # desc_number = dedent('''
    #                      Inserts IMGT numbering spacers into sequence_alignment, rebuilds the germline sequence
    #                      in germline_alignment, and adjusts the values in the coordinate fields v_germline_start
    #                      and v_germline_end accordingly.
    #                      ''')
    # parser_number = subparsers.add_parser('number', parents=[parser_parent],
    #                                       formatter_class=CommonHelpFormatter, add_help=False,
    #                                       help='Add IMGT-numbering to an AIRR Rearrangement TSV.',
    #                                       description=desc_number)
    # group_number = parser_number.add_argument_group('aligner parsing arguments')
    # group_number.add_argument('-i', nargs='+', action='store', dest='aligner_files', required=True,
    #                         help='''AIRR Rearrangement TSV files.''')
    # group_number.add_argument('-r', nargs='+', action='store', dest='repo', required=False,
    #                         help='''List of folders and/or fasta files containing
    #                              IMGT-numbered germline sequences corresponding to the
    #                              set of germlines used for the alignment.''')
    # parser_number.set_defaults(func=numberAIRR)

    return parser


if __name__ == "__main__":
    """
    Parses command line arguments and calls main
    """
    parser = getArgParser()
    checkArgs(parser)
    args = parser.parse_args()
    args_dict = parseCommonArgs(args, in_arg='aligner_files')

    # Set no ID parsing if sequence files are not provided
    if 'seq_files' in args_dict and not args_dict['seq_files']:
        args_dict['asis_id'] = True

    # Delete
    if 'aligner_files' in args_dict: del args_dict['aligner_files']
    if 'seq_files' in args_dict: del args_dict['seq_files']
    if 'cellranger_files' in args_dict: del args_dict['cellranger_files']
    if 'out_files' in args_dict: del args_dict['out_files']
    if 'command' in args_dict: del args_dict['command']
    if 'func' in args_dict: del args_dict['func']

    # Call main
    for i, f in enumerate(args.__dict__['aligner_files']):
        args_dict['aligner_file'] = f
        args_dict['out_file'] = args.__dict__['out_files'][i] \
                                if args.__dict__['out_files'] else None
        if 'seq_files' in args.__dict__:
            args_dict['seq_file'] = args.__dict__['seq_files'][i] \
                                    if args.__dict__['seq_files'] else None
        if 'cellranger_files' in args.__dict__:
            args_dict['cellranger_file'] = args.__dict__['cellranger_files'][i] \
                                           if args.__dict__['cellranger_files'] else None
        args.func(**args_dict)
